#include "GComboBox.h"


GComboBox::GComboBox()
{
  init();
  setActive(1);
  list = new GComboList(G_VERTICAL_SCROLL,this,x,y,width,height);
  button = new GDownButton(this,x+width-TEXT_HEIGHT,y,TEXT_HEIGHT,TEXT_HEIGHT+2);
}

GComboBox::GComboBox(int flags,GObject * parent,int x,int y,int width,int height)
{
  init(parent,NULL,x,y,width,TEXT_HEIGHT+2);
  setActive(1);
  list = new GComboList(flags,this,x,y+TEXT_HEIGHT+2,width,height);
  button = new GDownButton(this,x+width-TEXT_HEIGHT,y,TEXT_HEIGHT,TEXT_HEIGHT+2);
}

GComboBox::GComboBox(int flags,GObject * parent,int x,int y,int width,int thisHeight,int listHeight)
{
  init(parent,NULL,x,y,width,thisHeight);
  setActive(1);
  list = new GComboList(flags,this,x,y+TEXT_HEIGHT+2,width,listHeight);
  button = new GDownButton(this,x+width-8,y,8,thisHeight);
}

/*
  Remove the list and the button created in the consrtuctor
*/
void GComboBox::release()
{ 
  if(list){
    delete list;
    list = NULL;
  }
  if(button){
    delete button;
    button = NULL;
  }
}

GComboBox::~GComboBox()
{
  release();
}

/*
  set the back ground color of the ComboBox
*/
void GComboBox::bgColor()
{
  glColor3f(1.0f,1.0f,1.0f);
}

/*
  post: Draw the ComboBox.  If the item in the edit box
  is too long, then only draw the last part of it.  If
  This is selected then draw the dropdown list
*/
void GComboBox::draw()
{
  int textSize=getTextWidth(text);
  drawBackGround();
  
  if(textSize<this->getWidth()-button->getWidth()){
    drawForeGround();
  }
  else{
    int minSkipSize = textSize-this->getWidth()+button->getWidth();
    int i=0;
    int sizeSoFar=0;
    int len = strlen(text);
    char * tback=text;
    while(i<len && sizeSoFar < minSkipSize){
      sizeSoFar+=getCharWidth(text[i]);
      i++;		   
    }
    text=&text[i];
    drawForeGround();
    text=tback;
  }
  button->draw();
  
  if(list && state&G_SELECTED){
    glTranslatef(0,0,+0.1);
      list->draw();
      glTranslatef(0,0,-0.1);
  }
}

/*
  post: return the number of elements
*/
int GComboBox::getNumElements()
{
  return list->getNumElements();
}

/*
  post: set which list index is currently selected
*/
void GComboBox::setSelected(int i)
{
  list->setSelected(i);
  this->setText(list->getSelectedElement());
}

/*
  post: get which list index is currently selected
*/
int GComboBox::getSelected()
{
  return list->getSelected();
}

/*
  post: Get the selected element (pointer to char. this is not very safe)
*/ 
char * GComboBox::getSelectedElement()
{
  return list->getSelectedElement();
}

/* Signifies a change selection */
void GComboBox::defaultAction(GObject * go)
{
  if(go){
    if(list->getSelectedElement()>=0){
      this->setText(list->getSelectedElement());
    }
    this->state&=~G_SELECTED;
  }
}

/*
  post: remove all the elments from the drop down list.
        Then free this text and set to null
*/
void GComboBox::removeAllElements()
{
  if(list){
    list->removeAllElements();
    list->setWidth(this->getWidth());
  }
  if(text){
    free(text);
    text=NULL;
  }
}

/*
  post: add the element to the list.  If the edit box
  is not displaying anything, set its text to str
*/
void GComboBox::addElement(char * str)
{
  if(list && str){
    int textwidth = getTextWidth(str);
    list->addElement(str);
    if(list->getWidth()+20 < textwidth){
      list->setWidth(textwidth+20);
    }
    if(text==NULL){
      setText(list->getSelectedElement());
    }
  }
}

/*
  post: dispatch the moust pressed to the children
  Call the action iff the click hit the list
  but not hte list's scrollbar
*/
int GComboBox::mousePressed(int button,int x,int y)
{
  int ret=0;
  if(hit(x,y)){
    ret|=this->button->mousePressed(button,x,y);
    return 1;
  }
  else{
    if(state&G_SELECTED){
      if(list->mousePressed(button,x,y)){
	if(!list->hitScrollBar(x,y)){
	  if(action){
	    action(this);
	  }
	  else defaultAction(this);
	}
	ret=1;
      }
    }
  }
  return ret;
}

/*
  post: if this was hit, toggle selected.  If this
  is already selected then send the message to
  the list
*/
int GComboBox::mouseReleased(int button,int x,int y)
{
  int ret=0;
  if(hit(x,y)){
    ret|=this->button->mouseReleased(button,x,y);
    if(ret){
      state^=G_SELECTED;
      return 1;
    }
  }
  else{
    if(state&G_SELECTED){
      ret|=list->mouseReleased(button,x,y);
    }
  }
  return ret;
}

/*
  post: If this is selected forward the message
        to the list
*/
int GComboBox::mouseDragged(int x,int y)
{
  int ret=0;
  if(state&G_SELECTED){
    ret|=list->mouseDragged(x,y);
  }	
  return ret;
}

/*
  post: If this is selected forward the message
        to the list
*/
int GComboBox::mouseMoved(int x,int y)
{
  int ret=0;
  
  if(state&G_SELECTED){
    ret|=list->mouseMoved(x,y);
   }
  return ret;
}


