#include "GEditBox.h"


GEditBox::GEditBox()
{
  init();
  numLines=0;
  top=topi=0;
  cursori=0;
  maxTextSize=0;
  verticalScroll=horizontalScroll=NULL;
}

GEditBox::GEditBox(int flags, GObject * owner,char * str,int x,int y,int width,int height)
{
  init(owner,str,x,y,width,height);
  numLines=0;
  top=topi=0;
  cursori=0;
  maxTextSize=strlen(str);
  verticalScroll=horizontalScroll=NULL;
  if(flags&G_HORIZONTAL_SCROLL){
    horizontalScroll = new GScrollBar();
  }
  if(flags&G_VERTICAL_SCROLL){
    verticalScroll = 
      new GScrollBar(G_VERTICAL_SCROLL,this,x+width-TEXT_HEIGHT,y,TEXT_HEIGHT,height);
  }
}


void GEditBox::release()
{
  if(horizontalScroll){
    delete horizontalScroll;
    horizontalScroll=NULL;
  }
  if(verticalScroll){
    delete verticalScroll;
    verticalScroll=NULL;
  }
  maxTextSize=0;
  top=topi=0;
  numLines=0;
}

GEditBox::~GEditBox()
{
  release();
}

/*
  post:Draw the edit box.  Currently performes word wrapping
  on multiline edit boxes.
*/
void GEditBox::draw()
{
  drawBackGround();
  
  /*
    Draw the text in the window by calculating the
    width of each line.  THis currently wraps
    words in the middle of the word
  */
  int start=topi;
  int end=0;
  int len = strlen(text);
  int texty=TEXT_HEIGHT+2;
  int maxWidth = this->width;
  int cursorx=-1;
  int cursory=-1;
  char cback;
  if(verticalScroll)
    maxWidth-=verticalScroll->getWidth();
  fgColor();
  
  while(start<len && texty<height){
    end = getTextMaxSizeEx(text,start,maxWidth,cursori,&cursorx);
    if(cursorx!=-1 && cursory==-1){
      cursory=y+texty+2;
      cursorx+=(this->x+2);
    }
    cback=text[end];
    text[end]=0;
    glRasterPos2i(x+2,y+texty);
    drawText(&text[start]);
    text[end]=cback;
    start=end;
    texty+=TEXT_HEIGHT+2;
  }
  if(state&G_SELECTED){
    if(cursorx>=0 && cursory>=0){
      static int flash=0;
      if(flash&0x8){
	fgColor();
      }
      else bgColor();
      flash++;
      flash&=0xF;
      glBegin(GL_LINES);
      glVertex2i(cursorx,cursory);
      glVertex2i(cursorx,cursory-TEXT_HEIGHT);
      glEnd();
    }
  }
  if(verticalScroll)
    verticalScroll->draw();
  if(horizontalScroll)
    horizontalScroll->draw();

  //  glRasterPos2i(x,y+height+TEXT_HEIGHT);
  //drawText("%d,%d,%d",numLines,top,topi); 
}

/*
  post: calculates the number of lines in the text box.
  Should be called when text changes significantly
*/
void GEditBox::updateNumLines()
{
  int cursorx=0;
  int maxWidth = width;
  int len = strlen(text);
  int start=0;
  int end=0;
  if(verticalScroll)
    maxWidth-=verticalScroll->getWidth();
  numLines=0;
  while(start<len){
    end=getTextMaxSizeEx(text,start,maxWidth,cursori,&cursorx);
    numLines++;
    start=end;
  }
}

/*
  post: inserts the char c at the location where in the text
*/
void GEditBox::insertChar(int where,char c)
{
  int needLen = strlen(text)+1;
  int len = strlen(text);
  if(needLen>=maxTextSize){
    char * newText;
    
    while(needLen>=maxTextSize){
      maxTextSize=2*(maxTextSize+1);
    }
    newText = (char *)malloc(maxTextSize+1);
    if(newText){
      newText[0]=0;
      memcpy(newText,text,strlen(text)+1);
    }
    //Fixme
    else{
      
    }
  }
  int i=len+1;
  while(i>where){
    text[i]=text[i-1];
    i--;
  }
  text[i]=c;
  updateNumLines();
}

/*
  post: set the text of this object using append text
*/
void GEditBox::setText(char * str)
{
  if(this->text){
    this->text[0]=0;
  }
  if(str)
    appendText(str);
  cursori=top=topi=0;//Is this ok
}

/*
  post: append the text to what is already in the buffer, which
  is just this->text
*/
void GEditBox::appendText(char * str)
{
  int needLen = strlen(text)+strlen(str);
  if(needLen>=maxTextSize){
    char * newText;
    while(needLen>=maxTextSize){
      maxTextSize=2*(maxTextSize+1);
    }
    newText = (char *)malloc(maxTextSize+1);
    if(newText){
      newText[0]=0;
      memcpy(newText,text,strlen(text)+1);
      sprintf(newText,"%s%s",text,str);
      if(text)
	free(text);	
      text=newText;
      
    }
    //Fixme
    else{
    }
  }
  else{
    
    strcat(text,str);
  }
  updateNumLines();
}

/*
  post: handle the key pressed event 
*/
int GEditBox::keyTyped(unsigned int mod,char key)
{
  if(state&G_SELECTED){
    /* delete the character at before cursori */
    if(key==8){
      if(cursori>0){
	cursori--;
	int len = (int)(strlen(text)+1);
	int i=cursori;
	while(i<len){
	  text[i]=text[i+1];
	  i++;
	}
	updateNumLines();
      }
    }
    /* insert the character at cursori */
    else{
      char str[2];
      str[0]=key;
      str[1]=0;
      if(text[cursori]==0){
	appendText(str);
	cursori++;
      }
      else{
	insertChar(cursori,key);
	cursori++;
      }
    }
    return 1;
  }
  return 0;
}
/* 
   post: forward to scrollbars if this is hit
*/
int GEditBox::mouseReleased(int button,int x,int y)
{
  if(hit(x,y)){
    if(verticalScroll)
      verticalScroll->mouseReleased(button,x,y);
    if(horizontalScroll)
      horizontalScroll->mouseReleased(button,x,y);
  }
  return 0;
}

/* need to deselect this some other way, may not get this message */
int GEditBox::mousePressed(int button,int x,int y)
{
  if(hit(x,y)){
    int ret=0;
    state|=G_SELECTED;
    if(verticalScroll)
      ret|=verticalScroll->mousePressed(button,x,y);
    if(horizontalScroll)
      ret|=horizontalScroll->mousePressed(button,x,y);
    if(!ret){
      int linei = (y-this->y)/(TEXT_HEIGHT+2);
      int curi=topi;
      int cursorx=0;
      int maxWidth=this->getWidth();
      if(verticalScroll)
	maxWidth-=verticalScroll->getWidth();
      x-=this->x;
      while(linei>0){
	curi=getTextMaxSizeEx(text,curi,maxWidth,0,&cursorx);
	linei--;
      }
      curi=getTextMaxSizeEx(text,curi,x,0,&cursorx);
      cursori=curi;
      //Set the cursor position
      
    }
    return 1;
  }
  state&=~G_SELECTED;
  return 0;
}

/* 
   post: forward to scrollbars if this is hit
*/
int GEditBox::mouseDragged(int x,int y)
{
  if(hit(x,y)){
    if(verticalScroll)
      verticalScroll->mouseDragged(x,y);
    if(horizontalScroll)
      horizontalScroll->mouseDragged(x,y);	
    return 1;
  }
  return 0;
}

/* 
   post: forward to scrollbars if this is hit
*/
int GEditBox::mouseMoved(int x,int y)
{
  if(hit(x,y)){
    if(verticalScroll)
      verticalScroll->mouseMoved(x,y);
    if(horizontalScroll)
      horizontalScroll->mouseMoved(x,y);
  }
  return 0;
}

/*
  post: scroll up in the text by one line if possible
*/
void GEditBox::scrollUp()
{
  int cursorx=0;
  int maxWidth = width;
  if(verticalScroll)
    maxWidth-=verticalScroll->getWidth();
  
  if(top>0){
    top--;
    int line=0;
    topi=0;
    while(line<top){
      topi=getTextMaxSizeEx(text,topi,maxWidth,0,&cursorx);
      line++;
    }
  }
}

/*
  post: scroll down in the text by one line if possible
*/
void GEditBox::scrollDown()
{
  int cursorx=0;
  int maxWidth = width;
  if(verticalScroll)
    maxWidth-=verticalScroll->getWidth();
  
  if(top<numLines-(height/(TEXT_HEIGHT+2))){
    topi=getTextMaxSizeEx(text,topi,maxWidth,0,&cursorx);
    top++;
  }
}

int GEditBox::getScrollAreaSize(GObject * c)
{
  if(verticalScroll==c){
    return numLines*(TEXT_HEIGHT+2);
  }
  return height;
}

int GEditBox::getScrollPosition(GObject * c)
{
  if(verticalScroll==c){
    return top*(TEXT_HEIGHT+2);
  }
  return 0;
}

void GEditBox::scrollAction1(GObject * c)
{
  if(verticalScroll==c){
    scrollUp();
  }
}

void GEditBox::scrollAction2(GObject * c)
{
  if(verticalScroll==c){
    scrollDown();
  }
}

void GEditBox::scrollBarPressed(GObject * c,int x,int y)
{
  if(verticalScroll==c){
    int newtop = (y*numLines)/x;
    int line=0;
    int cursorx=0;
    int maxWidth = width;
    if(verticalScroll)
      maxWidth-=verticalScroll->getWidth();
    if(newtop>=numLines-(height/(TEXT_HEIGHT+2)))
      newtop=numLines-(height/(TEXT_HEIGHT+2));
    if(newtop<0)
      newtop=0;
    topi=0;
    top=newtop;
    while(line<top){
      topi=getTextMaxSizeEx(text,topi,maxWidth,0,&cursorx);
      line++;
    }
  }
}



void GSpinEditBox::release()
{
	GEditBox::release();
	if(up){
		delete up;
		up=NULL;
	}
	if(down){
		delete down;
		down=NULL;
	}
}

void upButtonPressed(GObject * go)
{
	if(go){
		((GSpinEditBox *)go->getParent())->incrementUp();
	}
}
void downButtonPressed(GObject * go)
{
	if(go){
		((GSpinEditBox *)go->getParent())->decrementDown();
	}
}

GSpinEditBox::~GSpinEditBox()
{
	release();
}

void GSpinEditBox::setValue(float val)
{
  if(val>max)
    val=max;
  if(val<min)
    val=min;
  this->value = val;
  updateText();
}

void GSpinEditBox::setIncrement(float i)
{
	increment=i;
}
void GSpinEditBox::setMin(float m)
{
	min=m;
}
void GSpinEditBox::setMax(float m)
{
	max=m;
}
void GSpinEditBox::updateText()
{
  
  if(displayAsInt){
    sprintf(text,"%d",(int)value);
  }
  else
    sprintf(text,"%6.4f",value);
}
void GSpinEditBox::incrementUp()
{
  setValue(value+increment);
}
void GSpinEditBox::decrementDown()
{
  setValue(value-increment);
}

void GSpinEditBox::draw()
{
	GEditBox::draw();
	if(up){
		up->draw();
	}
	if(down){
		down->draw();
	}
}

 int GSpinEditBox::mouseReleased(int button,int x,int y)
{
	int ret;
	if(up){
		ret|=up->mouseReleased(button,x,y);
	}
	if(down){
		ret|=down->mouseReleased(button,x,y);
	}
	return ret;
}
int GSpinEditBox::mousePressed(int button,int x,int y)
{
  int ret=0;
  if(up){
    ret|=up->mousePressed(button,x,y);
    if(action)
      action(this);
  }
  if(down){
    ret|=down->mousePressed(button,x,y);
    if(action)
      action(this);
  }
  if(!ret)
    ret|=GEditBox::mousePressed(button,x,y);
  return ret;
}
int GSpinEditBox::mouseDragged(int x,int y)
{
	return 0;
}
int GSpinEditBox::mouseMoved(int x,int y)
{
	int ret;
	if(up){
		ret|=up->mouseMoved(x,y);
	}
	if(down){
		ret|=down->mouseMoved(x,y);
	}
	ret|=GEditBox::mouseMoved(x,y);
	return ret;
}

