#ifndef G_EDITBOX_H
#define G_EDITBOX_H

#include "GScrollBar.h"
#include "GScrollable.h"

#define G_MULTILINE    0x400
#define G_READONLY     0x800
#define G_NUMBERSONLY 0x1000
#define G_AUTOVSCROLL 0x2000
#define G_AUTOHSCROLL 0x4000




class GEditBox : public GScrollable
{
protected:
	int maxTextSize;
	GScrollBar * verticalScroll;
	GScrollBar * horizontalScroll;

	int cursori;
	int topi;//Index into the string
	int top; //actual line number
	int numLines; //The number of lines in the textbox...must keep track of somehow...

public:
	GEditBox();
	GEditBox(int flags, GObject * owner,char * str,int x,int y,int width,int height);

	virtual void release();
	virtual ~GEditBox();

	virtual void draw();

	virtual void bgColor(){glColor3f(1.0f,1.0f,1.0f);}
	virtual int flipShade(){return 1;}

	virtual void updateNumLines();
	virtual void setText(char * str);
	virtual void appendText(char * str);
	virtual void insertChar(int where,char c);
	virtual void clearText(){if(text)text[0]=0;}

	virtual int keyTyped(unsigned int mod,char key);
	virtual int mouseReleased(int button,int x,int y);
	virtual int mousePressed(int button,int x,int y);
	virtual int mouseDragged(int x,int y);
	virtual int mouseMoved(int x,int y);

	virtual void scrollUp();
	virtual void scrollDown();

	virtual int getScrollAreaSize(GObject *);
	virtual int getScrollPosition(GObject *);
	virtual void scrollAction1(GObject *);
	virtual void scrollAction2(GObject *);
	virtual void scrollBarPressed(GObject *,int x,int y);
};

void upButtonPressed(GObject * go);
void downButtonPressed(GObject * go);

class GSpinEditBox : public GEditBox
{
	
 public:
  class GUpButton : public GButton
    {
    public:
      GUpButton(GObject  * owner,char * str,int x,int y,int width,int height):GButton(owner,str,x,y,width,height)
	{
	  
	}
      void draw(){
	drawBackGround();
	fgColor();
	glBegin(GL_TRIANGLES);
	glVertex2i(x+width/2,y);
	glVertex2i(x,y+height);
	glVertex2i(x+width,y+height);
	glEnd();
      }
      void defaultAction(GObject * go)
	{
	  ((GSpinEditBox *)getParent())->incrementUp();
	}
    };
  class GDownButton : public GButton
    {
    public:
      GDownButton(GObject  * owner,char * str,int x,int y,int width,int height):GButton(owner,str,x,y,width,height)
	{
	  
	}
      void draw(){
	drawBackGround();
	fgColor();
	glBegin(GL_TRIANGLES);
	glVertex2i(x,y);
	glVertex2i(x+width,y);
	glVertex2i(x+width/2,y+height);
	glEnd();
      }
      void defaultAction(GObject * go)
	{
	  ((GSpinEditBox *)getParent())->decrementDown();
	}
    };
  GSpinEditBox(): GEditBox()
    {
      value=0;
      min=0;
      max=1.0f;
      increment=0.1;
      displayAsInt=0;
      up = new GButton(this,"^",x+width-8,y,8,8);
      down = new GButton(this,"v",x+width-8,y+height-8,8,8);
      text = (char *)malloc(20);
	
      updateText();
    }
  GSpinEditBox(int flags, GObject * owner,float val,int x,int y,int width,int height) : GEditBox(flags,owner,"",x,y,width,height)
    {
      free(text);
      text= (char *) malloc(20);
      value=val;
      min=0;
      max=1.0f;
      increment=0.1;
      displayAsInt=0;
      up = new GUpButton(this,"^",x+width-height/2,y,height/2,height/2);
      down = new GDownButton(this,"v",x+width-height/2,y+height-height/2,height/2,height/2);
      
      updateText();
    }
  virtual void release();
  virtual ~GSpinEditBox();
  
  float getValue() {return value;}
  void setValue(float value);
  void setIncrement(float i);
  void setMin(float m);
  void setMax(float m);
  void updateText();
  void incrementUp();
  void decrementDown();
  void setDisplayAsInt(int i){displayAsInt=1;}
  
  virtual void draw();
  virtual int mouseReleased(int button,int x,int y);
  virtual int mousePressed(int button,int x,int y);
  virtual int mouseDragged(int x,int y);
  virtual int mouseMoved(int x,int y);
	
  
protected:
  GButton * up;
  GButton * down;
  float min;
  float max;
  float value;
  float increment;
  int displayAsInt;
};
#endif


