#ifndef G_FILEWINDOW
#define G_FILEWINDOW

#ifdef _WIN32
#include <windows.h>
#endif

#include "GWindow.h"
#include "GComboBox.h"
#include "GList.h"
#include "GEditBox.h"

#include <sys/stat.h>
#include <dirent.h>
#include <unistd.h>
#include "GFileFilter.h"
#include <ctype.h>
class GFileWindow: public GWindow
{
protected:
  int selectDirs;
  GComboBox * dirCB;
  GList     * dirList;
  GList     * fileList;  
  GEditBox * fileName;
  GComboBox * fileFilterCB;
  GButton * ok;
  GButton * cancel;
 public:
  static int numActive;
  int    result;
public:
  GFileFilter ** fileFilters;
  int  numFileFilters;

  /*
    okButtonAction(GObject * go)
    post: call the callback if it is set with the
    value and the path name
  */
  static void okButtonAction(GObject * go)
  {
    if(go && go->getParent()){
      char relFileName[1024];
      char path[1024];
      char * fullPath;
      GFileWindow  * gf = (GFileWindow *) go->getParent();
      
      gf->fileName->getText(relFileName,1024);
      getcwd(path,1024);
      
      fullPath = new char[strlen(relFileName)+strlen(path)+2];
      sprintf(fullPath,"%s%c%s",path,
	      DIR_SEPARATOR,
	      relFileName);
      printf("relFileName %s\n",relFileName);
      gf->setShowing(0);
      if(gf->fileWindowCallback)
	gf->fileWindowCallback(1,fullPath);
      else{
	printf("selected: %s\n",fullPath);
	delete [] fullPath;
      }
      gf->postDestroy();
    }
  }
  /*
    cancelButtonAction
    pre:
    post: if the action callback is set call it with a value
    of 0 and the fullPath name
  */
  static void cancelButtonAction(GObject * go)
  {
    if(go && go->getParent()){
      char relFileName[1024];
      char path[1024];
      char * fullPath;
      GFileWindow  * gf = (GFileWindow *) go->getParent();
      
      gf->fileName->getText(relFileName,1024);
      getcwd(path,1024);
      
      fullPath = new char[strlen(relFileName)+strlen(path)+2];
      sprintf(fullPath,"%s%c%s",path,
	      DIR_SEPARATOR,
	      relFileName);

      gf->setShowing(0);
      if(gf->fileWindowCallback)
	gf->fileWindowCallback(0,fullPath);
      else{
	delete [] fullPath;
      }
      gf->postDestroy();
    }   
  }

  /*
    post:
    update the current directory to 
  */
  static void dirCBAction(GObject * go)
  {
    if(go && go->getParent()){
      char buffer[1024];
      go->defaultAction(go);
      GFileWindow  * gf = (GFileWindow *) go->getParent();
      ((GList *) go)->getText(buffer,256);
      gf->setCurrentDirectory(buffer);
    }
  }
  /*
    post: reread the directory with the new file filter
  */
  static void fileFilterCBAction(GObject * go){
    if(go && go->getParent()){
      go->defaultAction(go);
      GFileWindow * gf = (GFileWindow *) go->getParent();
      gf->changedDir();
    }
  }
  /*
    post: change the directory to the selected directory 
  */
  static void dirListAction(GObject * go)
  {
    if(go && go->getParent()){
      char buffer[1024];
      GFileWindow  * gf = (GFileWindow *) go->getParent();
      sprintf(buffer,((GList *) go)->getSelectedElement());
      if(strstr(buffer,"..")){
	gf->goUpDir();
      }
      else{
	gf->setCurrentDirectory(buffer);
      }
    }
  }
  static void fileListAction(GObject * go)
  {
    /* Fixme : this is ugly */
    if(go && go->getParent()){
      GFileWindow  * gf = (GFileWindow *) go->getParent();
      printf("%s\n",((GList *) go)->getSelectedElement());
      (gf->fileName)->setText(((GList *) go)->getSelectedElement());
    }
  }
public:

  GFileWindow::~GFileWindow(){
    release();
  }
  /*
    post: set the current working directory back to what it was
  */
  void release(){
    if(fileFilters){
      int i=0;
      while(i<numFileFilters){
	delete fileFilters[i];
	fileFilters[i]=NULL;
	i++;
      }
      delete fileFilters;
    }
    GFileWindow::numActive--;
  }

  
  GFileWindow(GFileFilter ** filters=NULL, int numFilters=0)
    :GWindow(NULL,"File",10,10,260,210){
    GFileWindow::numActive++;
    fileFilters=filters;
    this->numFileFilters=numFilters;
    if(filters==NULL){
      fileFilters = new (GFileFilter *)[2];
      fileFilters[1]=new GFileFilter();
      fileFilters[0]=new GFileFilterDemoDir();
      numFileFilters=2;
    }
    fileWindowCallback=0;
    destroyOnClose();
    selectDirs=1;
    addObject(new GStatic(this,"Path:",10,20,30,TEXT_HEIGHT+2));
    dirCB = new GComboBox(G_VERTICAL_SCROLL,this,40,20,140,TEXT_HEIGHT+4,60);
    
    //addObject(new GStatic(this,"Directories",30,40,60,TEXT_HEIGHT+2));
    //addObject(new GStatic(this,"Open",150,40,60,TEXT_HEIGHT+2));
    dirList = new GList(G_HORIZONTAL_SCROLL|
			G_VERTICAL_SCROLL,this,10,20+TEXT_HEIGHT+6,115,110);
    fileList = new GList(G_HORIZONTAL_SCROLL|
			 G_VERTICAL_SCROLL,this,130,20+TEXT_HEIGHT+6,115,110);

    ok=new GButton(this,"Open",
			  175,144+TEXT_HEIGHT,60,TEXT_HEIGHT+4);
    addObject(ok);
    
    cancel=new GButton(this,"Cancel",
			  175,144+2*(TEXT_HEIGHT+4),60,TEXT_HEIGHT+4);
    addObject(cancel);
  
    addObject(new GStatic(this," Filename:",
			  10,144+TEXT_HEIGHT,60,TEXT_HEIGHT+4));
    addObject(new GStatic(this,"File Type:",
			  10,144+2*(TEXT_HEIGHT+4),60,TEXT_HEIGHT+2));
    fileName = new GEditBox(0,this,"filename.txt",
			    70,144+TEXT_HEIGHT,100,TEXT_HEIGHT+4);
    fileFilterCB = new GComboBox(0,this,70,144+2*(TEXT_HEIGHT+4)
			       ,100,TEXT_HEIGHT+2,40);
    int i=0;
    while(i<numFileFilters){
      fileFilterCB->addElement(fileFilters[i]->getName());
      i++;
    }
    
    dirCB->action = GFileWindow::dirCBAction;
    dirList->doubleClickAction = GFileWindow::dirListAction;
    fileList->action = GFileWindow::fileListAction;
    fileFilterCB->action = GFileWindow::fileFilterCBAction;

    ok->action = GFileWindow::okButtonAction;
    cancel->action = GFileWindow::cancelButtonAction;
    addObject(fileName);
    addObject(fileFilterCB);
    addObject(fileList);
    addObject(dirList);
    addObject(dirCB);
    currentDir[0]=0;
    getcwd(currentDir,256);
    changedDir();
  }
  void goUpDir(){
    chdir("..");
    changedDir();
  }
  void setCurrentDirectory(char * dir){
    chdir(dir);
    changedDir();
  }
  int fileIsAccepted(char * file){
    int selectedff = fileFilterCB->getSelected();
    if(selectedff>=0 && selectedff<numFileFilters){
      return fileFilters[selectedff]->accept(file);
    }
    return 0;
  }
  /*
    void changedDir
    post: update the dialog to correspond to the current directory
  */
  void changedDir(){
    DIR * dir = opendir("./");
    struct stat  buf;
    struct dirent * de;
    char str[255];
    char * ptr;
    char c;
    if(!dir){  
      return;
    }
    de=readdir(dir);
    getcwd(str,255);

    dirCB->removeAllElements();
    ptr=str;
#ifdef _WIN32
    ptr[0]=toupper(ptr[0]);//Take the drive string to upper case
    char driveString[256];
    char * driveStringPtr=driveString;
    int dstri=0;
    int dstrlen = GetLogicalDriveStrings(256,driveString);
    while(dstri<dstrlen){
      if(driveStringPtr[0]==0)
	break;
      driveStringPtr[0]=toupper(driveStringPtr[0]);
      if(driveStringPtr[0]==ptr[0]){
	driveStringPtr+=(strlen(driveStringPtr)+1);
	dstri+=(strlen(driveStringPtr)+1);
	break;
      }
      dirCB->addElement(driveStringPtr);
      driveStringPtr+=(strlen(driveStringPtr)+1);
      dstri+=(strlen(driveStringPtr)+1);
    }
#endif
    /*
      This takes care of the case when the current dir
      is not the root, so we want to add the root
      as the first element in the combo box.
      If it is the root, it will get added later
      anyhow.
    */
    if((ptr=strchr(str,DIR_SEPARATOR))){
      if(ptr[1]!=0){
	c=ptr[1];
	ptr[1]=0;
	dirCB->addElement(str);
	ptr[1]=c;
      }
    }
    ptr=str;    
    ptr++;
#ifdef _WIN32
    ptr+=2;//Leave enough room for the first slash ..ie c:
#endif
    while((ptr=strchr(ptr,DIR_SEPARATOR))){
      c=ptr[0];
      ptr[0]=0;
      dirCB->addElement(str);
      ptr[0]=c;
      ptr++;
    }
    dirCB->addElement(str);
  
    dirCB->setSelected(dirCB->getNumElements()-1);
    /*
      Add the remaining drive letters to the 
      dir combo list
    */
#ifdef _WIN32
    while(dstri<dstrlen){
      driveStringPtr[0]=toupper(driveStringPtr[0]);
      if(driveStringPtr[0]==0)
	break;
      dirCB->addElement(driveStringPtr);
      driveStringPtr+=(strlen(driveStringPtr)+1);
      dstri+=(strlen(driveStringPtr)+1);
    }
#endif
    fileList->removeAllElements();
    dirList->removeAllElements();
    while(de){
      int error;
      if((error=stat(de->d_name,&buf))<0){
	printf("error stating %s %d\n",de->d_name,error);
	perror("error:");
	break;
      }
      else if(S_ISREG(buf.st_mode)){
	if(fileIsAccepted(de->d_name)){
	  fileList->addElement(de->d_name);
	}
      }
      else if(S_ISDIR(buf.st_mode)){
	dirList->addElement(de->d_name);
	if(selectDirs){
	  //if(!(de->d_name[0]=='.' && (de->d_name[1]==0||de->d_name[1]=='.')))
	  {
	    if(fileIsAccepted(de->d_name))
	      fileList->addElement(de->d_name);
	  }
	}
      }
      de=readdir(dir);
    }
    fileName->setText(fileList->getSelectedElement());
    closedir(dir);
  }
  void setShowing(int i)
  {
    GWindow::setShowing(i);
    if(!i){
      chdir(currentDir);
    }
  }
  void (* fileWindowCallback)(int, char *);
protected:
  char currentDir[1024];
};


#endif





