#include "GList.h"


GList::GList()
{
  init();
  top=0;
  left=0;
  selected=0;
  initElements();
  verticalScroll=horizontalScroll=NULL;
  doubleClickAction=NULL;
}

GList::GList(int flags,GObject * parent, int x,int y,int width,int height)
{
  init(parent,"",x,y,width,height);
  doubleClickAction=NULL;
  top=0;
  left=0;
  selected=0;
  textWidth=0;
  initElements();
  state|=flags;
  verticalScroll=horizontalScroll=NULL;
  if(state&G_HORIZONTAL_SCROLL){
    horizontalScroll = new GScrollBar(G_HORIZONTAL_SCROLL,
				      this,x,y+height-TEXT_HEIGHT,
				      width,TEXT_HEIGHT);//FIXME:
  }
  if(state&G_VERTICAL_SCROLL){
    verticalScroll = new GScrollBar(G_VERTICAL_SCROLL,this,x+width-TEXT_HEIGHT,y,TEXT_HEIGHT,height);
  }
}

/*
  post: initzlize the number of elements, should be called only once
*/
void GList::initElements()
{
  numElements=0;
  maxElements=GLIST_MAX_ELEMENTS;
  elements = (char **)malloc(sizeof(char *)*maxElements);
  if(!elements){
    //FIXME: error
  }
  memset(elements,0,sizeof(char *)*maxElements);
}

/*
  post: free each element, free the scroll bars
  if there are any
*/
void GList::release()
{
  if(elements){
    int i=0;
    while(i<numElements){
      if(elements[i])
	free(elements[i]);
      i++;
    }
    free(elements);
    elements=NULL;
  }
  if(verticalScroll){
    delete verticalScroll;
    verticalScroll =NULL;
  }
  if(horizontalScroll){
    delete horizontalScroll;
    horizontalScroll=NULL;
  }
  top=left=0;
}

GList::~GList()
{
  release();
}

/*
  post: draw the list, using the top index of topi
*/
void GList::draw()
{
  glColor3f(1.0f,1.0f,1.0f);
  glBegin(GL_QUADS);
  glVertex2i(x,y);
  glVertex2i(x+width,y);
  glVertex2i(x+width,y+height);
  glVertex2i(x,y+height);
  glEnd();

  int drawHScroll=(width<textWidth &&horizontalScroll);
  int drawVScroll=0;
  int i=top;
  int bottom = numElements;
  int maxHeight = height;
  int maxWidth = this->width;
  int elHeight = TEXT_HEIGHT+2;
  if(drawHScroll)maxHeight-=elHeight;
  drawVScroll = (numElements>(maxHeight/elHeight))&& (verticalScroll!=NULL);
  if((maxHeight/elHeight)+top<bottom){
    bottom = (maxHeight/elHeight)+top;
  }
  if(verticalScroll){
    maxWidth-=verticalScroll->getWidth();
  }else drawVScroll=0;
  while(i<bottom){
    if(i==selected){
      glColor3f(0.6f,0.6f,1.0f);
      glBegin(GL_QUADS);
      glVertex2i(x,y+elHeight*(i-top));
      glVertex2i(x+width,y+elHeight*(i-top));
      glVertex2i(x+width,y+elHeight*(i-top+1));
      glVertex2i(x,y+elHeight*(i-top+1));
      glEnd();
    }
    fgColor();
    glRasterPos2i(x,y+elHeight*(i-top+1));
    drawTextFromTo(elements[i],left,left+this->width);
    i++;
  }
  if(verticalScroll){
    if(drawHScroll){
      verticalScroll->setHeight(height-horizontalScroll->getHeight());
    }
    else verticalScroll->setHeight(height);
    if(drawVScroll)
      verticalScroll->draw();
  }
  if(horizontalScroll){
    if(drawVScroll){
      horizontalScroll->setWidth(width-verticalScroll->getWidth());
      
    }else
      horizontalScroll->setWidth(width);
    if(drawHScroll)
      horizontalScroll->draw();
  }
  if(drawHScroll && drawVScroll){
    int sx,sy,wx,wy;
    sy=horizontalScroll->getYPosition();
    sx=verticalScroll->getXPosition();
    wx=verticalScroll->getWidth();
    wy=horizontalScroll->getHeight();
    bgColor();
    glBegin(GL_QUADS);
    glVertex2i(sx,sy);
    glVertex2i(sx,sy+wy);
    glVertex2i(sx+wx,sy+wy);
    glVertex2i(sx+wx,sy);
    glEnd();
  }
}

/*
  post: set width, must account for scrollbars if any
*/
void GList::setWidth(int width)
{
  GObject::setWidth(width);
  if(verticalScroll){
    verticalScroll->setXPosition(getXPosition()+getWidth()
				 -verticalScroll->getWidth());
  }  
}

/*
  post: set the selected index to be i
*/
void GList::setSelected(int i)
{
  if(i>=0 && i<numElements){
    selected =i;
  }
}

/*
  post: return how many elements are in the list
*/
int GList::getNumElements()
{
  return numElements;
}

/*
  post: retun which element is selected.
  returns -1 if no elements are in the list
*/
int GList::getSelected()
{
  if(numElements && selected < numElements){
    return selected;
  }
  return -1;
}

/*
  post: returns the text of the selected element
  returns null if there is no elements in the list
*/
char * GList::getSelectedElement()
{
  if(numElements && selected < numElements){
    return elements[selected];
  }
  return NULL;
}

/*
  post: free and remove all elements from the list
*/
void GList::removeAllElements()
{
  int i=0;
  while(i<numElements){
    if(elements[i]){
      free(elements[i]);
      elements[i]=NULL;
    }
    i++;
  }
  left=0;
  textWidth=0;
  top=0;
  selected=0;
  numElements=0;
}

/*
  post: the string str is copied into the next list location
*/
void GList::addElement(char * str)
{
  if(numElements<maxElements){
    elements[numElements] = (char *) malloc(strlen(str)+1);
    sprintf(elements[numElements],str);
    if(getTextWidth(str)>textWidth){
      textWidth = getTextWidth(str);
    }
    numElements++;
  }
  else{
    //FIXME:
  }
}

/*
  post: return one if the x,y hits the scrollbars
*/
int GList::hitScrollBar(int x,int y)
{
  if(hit(x,y)){
    if(verticalScroll && verticalScroll->hit(x,y)){
      return 1;
    }
    if(horizontalScroll && horizontalScroll->hit(x,y)){
      return 1;
    }
  }
  return 0;
}

/*
  post: if this is hit, change the currently selected
  elemenent.  This calls the action. If the currently
  selected elmeent is the same as the new one
  the doubleClickAction is called
*/
int GList::mousePressed(int button,int x,int y)
{
  if(hit(x,y)){
    int curSelected = selected;
    
    if(verticalScroll){
      if(verticalScroll->mousePressed(button,x,y))
	return 1;
    }
    if(horizontalScroll){
      if(horizontalScroll->mousePressed(button,x,y))
	return 1;
    }
    /* FIXME: should this be in the default action? */
    int elHeight = TEXT_HEIGHT + 2;
    selected = ((y-this->y)/elHeight)+top;
    if(selected>=numElements){
      selected=numElements-1;
    }
    if(selected<0)
      selected=0;
    if((curSelected==selected) && doubleClickAction){
      doubleClickAction(this);
    }
    if(action){
      action(this);
    }
    return 1;
    
  }
  return 0;
}

/*
  post: forward the mouse released event
*/
int GList::mouseReleased(int button,int x,int y)
{
  if(hit(x,y)){
    if(verticalScroll){
      if(verticalScroll->mouseReleased(button,x,y))
	return 1;
    }
    if(horizontalScroll){
      if(horizontalScroll->mouseReleased(button,x,y))
	return 1;
    }
    return 1;
  }
  return 0;
}

/*
  post: forward the mouse dragged event
*/
int GList::mouseDragged(int x,int y)
{
  if(hit(x,y)){
    if(verticalScroll){
      if(verticalScroll->mouseDragged(x,y)){
	return 1;
      }
    }
    if(horizontalScroll){
      if(horizontalScroll->mouseDragged(x,y)){
	return 1;
      }
    }
    return 1;
  }
  return 0;
}

/*-----Scrollable interface--------*/
/*
  post: return the size of the scroll are
*/ 
int GList::getScrollAreaSize(GObject * caller)
{
  if(caller==verticalScroll){
    int size = height;
    if(size<numElements*(TEXT_HEIGHT+2))
      size = numElements*(TEXT_HEIGHT+2);
    return size;
  }
  else{

    //FIXME: horizontal
    return textWidth;
  }
}

/*
  post: get the current location in the scroll position
*/
int GList::getScrollPosition(GObject * caller)
{
  if(caller==verticalScroll){
    return top*(TEXT_HEIGHT+2);
  }
  else {
    return left;
  }
  //FIXME: horizontal
  return height;
}

/*
  post: this is the up action, so scroll up
*/
void GList::scrollAction1(GObject * caller)
{
  if(caller == verticalScroll){
    top--;
    if(top<0)
      top=0;
  }
  else if(caller==horizontalScroll){
    left-=TEXT_HEIGHT/2;
    if(left<0)
      left=0;
  }
}

/*
  post: this is the down action, so scroll down
*/
void GList::scrollAction2(GObject * caller)
{
  if(caller == verticalScroll){
    int elHeight = TEXT_HEIGHT+2;
    top++;
    if(top>=(numElements-(height/elHeight)))
      top=numElements-(height/elHeight);
    if(top<0)
      top=0;
  }
  else if(caller==horizontalScroll){
    left+=TEXT_HEIGHT/2;
    if(left>=textWidth-width)
      left = textWidth-width;
  }
}

/*
  post: set the selected element to be x/y units down the list
*/
void GList::scrollBarPressed(GObject * caller,int x,int y)
{
  if(caller==verticalScroll){
    int elHeight = TEXT_HEIGHT+2;
    if(x!=0){
      top = (numElements*y)/x;
      if(top>=numElements-height/elHeight)
	top = numElements-height/elHeight;
      if(top<0)
	top=0;
    }
  }
}


