#include "GMenu.h"

GObject * menuItems;
int  numMenuItems;


GMenuItem::GMenuItem()
{
  next=NULL;
  init();
}

GMenuItem::GMenuItem(GObject * parent,char * str,int x,int y,int width,int height)
{
  next=NULL;
  if(str[0]==0)
    height=4;
  if(width<getTextWidth(str)){
    width=getTextWidth(str)+4;
  }
  init(parent,str,x,y,width,height);
}


/*
  post: this does the shading when the mouse is over
        the menuitem
*/
int GMenuItem::flipShade()
{
  if(isSeparator())
    return 0;
  return GObject::flipShade();
}

/*
  post:Draw the menu item, do all the shading of hte
  sides in this call
*/
void GMenuItem::draw()
{
  if(state&G_MOUSEOVER && !isSeparator())
    fgColor();
  else bgColor();
  
  glBegin(GL_QUADS);
   glVertex2i(x,y);
   glVertex2i(x+width,y);
   glVertex2i(x+width,y+height);
   glVertex2i(x,y+height);
  glEnd();

  shadeHiColor();
  glBegin(GL_LINES);
   glVertex2i(x,y);
   glVertex2i(x,y+height);
  glEnd();
  shadeLoColor();
  glBegin(GL_LINES);
   glVertex2i(x+width,y);
   glVertex2i(x+width,y+height);
  glEnd();

  if(state&G_MOUSEOVER && !isSeparator())
    bgColor();
  else fgColor();
	
  glRasterPos2i(x+2,y+TEXT_HEIGHT);
  if(text==NULL || strcmp(text,"")!=0){
    drawText(text);
  }
  else{
    glBegin(GL_LINES);
    glVertex2i(x+2,y+height/2);
    glVertex2i(x+width-4,y+height/2);
    glEnd();
  }

}

/*
  post: make the mouse dragged look like the mouse moved
*/
int GMenuItem::mouseDragged(int x,int y)
{
  return mouseMoved(x,y);
}


GMenu::GMenu()
{
  init();
  initMenuItems();
  state|=MENU_VERTICAL;
  next=NULL;
}

GMenu::GMenu(int type,GObject * parent,char * str,int x,int y,int width,int height)
{
  init(parent,str,x,y,width,height);
  initMenuItems();
  state|=type;
  next=NULL;
}

void GMenu::initMenuItems()
{
  menuItems=NULL;
}

GMenu::~GMenu()
{
  release();
}

void GMenu::release()
{
  GMenuItem * mi = menuItems;
  while(mi){
    GMenuItem * oldMi = mi;
    mi = mi->next;
    delete oldMi;
  }
  menuItems=NULL;
}
	
void GMenu::draw()
{
  if(state&MENU_POPUP){
    drawBackGround();
    drawForeGround();
    drawHiShade();
    shadeLoColor();
    glBegin(GL_LINES);
     glVertex2i(x+width,y);
     glVertex2i(x+width,y+height);
    glEnd();
    GMenuItem * mi = menuItems;
    
    while(mi){
      mi->draw();
      mi = mi->next;
    }
  }
  else{
    drawBackGround();
    if(state&(G_MOUSEOVER|G_SELECTED)){
      drawHiShade();
      drawLoShade();
    }
    drawForeGround();
    if(state&G_SELECTED){
      GMenuItem * mi = menuItems;
      while(mi){
	mi->draw();
	mi = mi->next;
      }
    }
  }
}
/*
  post: add the menu item to the menu, setting it size
  and location as required by the type of menu
*/
void GMenu::addMenuItem(GMenuItem * item)
{
  GMenuItem * finger = menuItems;
  
  if(!item)
    return;
  if(state&MENU_VERTICAL){
    if(!finger){
      menuItems = item;
      item->setXPosition(x);
      item->setYPosition(this->y+this->height);
      if(!item->isSeparator())	
	item->setHeight(height+2);
      if(state&MENU_POPUP){
	if(item->getWidth()>getWidth())
	  this->setWidth(item->getWidth());
	else item->setWidth(getWidth());
      }
    }
    else{
      
      while(finger->next){
	if(item->getWidth()>finger->getWidth())
	  finger->setWidth(item->getWidth());
	else item->setWidth(finger->getWidth());
	finger = finger->next;
      }
      if(item->getWidth()>finger->getWidth())
	finger->setWidth(item->getWidth());
      else item->setWidth(finger->getWidth());
      
      item->setXPosition(x);
      item->setYPosition(finger->getYPosition()+finger->getHeight());
      if(!item->isSeparator())
	item->setHeight(height+2);
      
      if(state&MENU_POPUP){
	this->setWidth(item->getWidth());
      }
      finger->next = item;
    }
  }
  item->setParent(this);
}

/*
  post: a mouse pressed on a normal menu should toggle 
  whether it is selected or not.
  If the item is selected, then the pressed event is
  sent to the children
*/
int GMenu::mousePressed(int button,int x,int y)
{
  int ret=0;
  if(hit(x,y)){
    if(!(this->state&MENU_POPUP))
      this->state^=G_SELECTED;
    return 1;
  }
  /*
    ADDED the first choice july 14
    If the mouse isn't currently selected, but this hits 
    one of the children, it should be selected
   
  */
  else if(this->state&MENU_POPUP && isShowing()
	  ||this->state & G_SELECTED){
    GMenuItem * finger = menuItems;
    while(finger){
      if(finger->mousePressed(button,x,y)){
	ret|=1;
	break;
      }
      finger = finger->next;
    }
  }
  if(!ret){
    this->state&=~G_SELECTED;
  }
  return ret;
}

/*
  post: if this is selected or this is a popup, 
  then send the mousereleased to the children
*/
int GMenu::mouseReleased(int button,int x,int y)
{
  int ret=0;
  GMenuItem * finger = menuItems;
  if(this->state & G_SELECTED || state&MENU_POPUP){
    while(finger){
      if(finger->mouseReleased(button,x,y)){
	ret=1;
	break;
      }
      finger = finger->next;
    }
  }
  if(ret){
    state&=~G_SELECTED;
  }
  return ret;
}

/*
  post: dispatch the mouseMoved to the children
*/
int GMenu::mouseMoved(int x,int y)
{
  int ret=GObject::mouseMoved(x,y);
  if(this->state & G_SELECTED || state&MENU_POPUP){
    GMenuItem * finger = menuItems;
    while(finger){
      if(finger->mouseMoved(x,y)){
	ret=1;
      }
      finger = finger->next;
    }
    return ret;
  }
  return ret;
}



