#ifndef G_MENU_H
#define G_MENU_H

#include "GObject.h"


#define MENU_VERTICAL   0x100
#define MENU_HORIZONTAL 0x200
#define MENU_POPUP      0x400

class GMenuItem : public GObject
{
public:
	GMenuItem * next;
	GMenuItem();
	GMenuItem(GObject * parent,char * str,int x,int y,int width,int height);
	virtual ~GMenuItem(){};
	virtual void draw();
	virtual int flipShade();
	virtual int mouseDragged(int x,int y);
	virtual int isSeparator() {if(text && text[0]!=0)return 0;else return 1;}
};

class GMenu : public GMenuItem
{
protected:
	GMenuItem * menuItems;
	int  numMenuItems;
public:
	GMenu * next;
	GMenu();
	GMenu(int type,GObject * parent,char * str,int x,int y,int width,int height);

	virtual ~GMenu();

	virtual int mousePressed(int button,int x,int y);
	virtual int mouseReleased(int button,int x,int y);
	virtual int mouseMoved(int x,int y);

	virtual void release();
	
	virtual void draw();
	virtual void initMenuItems();
	virtual void addMenuItem(GMenuItem * );

};

class GMenuBar : public GObject
{
protected:
  GMenu * menus;
public:
  GMenuBar() {init();menus=NULL;state=0;};
  GMenuBar(GObject * parent,int x,int y,int width,int height) {init(parent,"",x,y,width,height);state=0;menus=NULL;};
  virtual void release(){
      GMenu * f = menus;
      while(f){
	GMenu * old = f;
	f=f->next;
	delete old;
      }
      menus=NULL;
  };
  virtual void draw(){
    drawBackGround();
    GMenu * m = menus;
    while(m){
      m->draw();
      m =m->next;
    }
  };
  virtual int mousePressed(int state, int x,int y){
    GMenu * m = menus;
    while(m){
      if(m->mousePressed(state,x,y))
	return 1;
      m =m->next;
    }
    return 0;
  }
  virtual int mouseMoved(int x,int y){
    int ret=0;
    GMenu * m = menus;
    while(m){
      if(m->mouseMoved(x,y))
	ret=1;
      m =m->next;
    }
    return ret;
  }
  virtual int mouseDragged(int x,int y){
    int ret=0;
    GMenu * m = menus;
    while(m){
      if(m->mouseDragged(x,y))
	ret=1;
      m =m->next;
    }
    return ret;
  }
  virtual int mouseReleased(int state,int x,int y){
      int ret=0;
      GMenu * m = menus;
      while(m){
	if(m->mouseReleased(state,x,y))
	  ret=1;
	m =m->next;
      }
      return ret;
  }

  
  void addMenu(GMenu * m){
    if(menus){
      GMenu * finger = menus;
      while(finger->next)finger=finger->next;
      finger->next=m;
      m->setXPosition(finger->getXPosition()+finger->getWidth());
      m->setYPosition(this->y+4);//FIXME
    }
    else{
      menus = m;
      m->setXPosition(this->x+4);
      m->setYPosition(this->y+4);
    }
  }
};

#endif

