#include "GObject.h"

#include <stdarg.h> 

void drawText(const char * fmt,...)
{
  char		text[1024];								
  va_list		ap;									
  
  if (fmt == NULL)									
    return;										       
  va_start(ap, fmt);									
  vsprintf(text, fmt, ap);						
  va_end(ap);											
  
  unsigned int i=0;
  while(i<strlen(text)){
    glutBitmapCharacter(GLUT_BITMAP_HELVETICA_10 , text[i]);
    i++;
  }
}

int getTextMaxSizeEx(char * str,int offset,int maxSize,int speciali,int * specialx)
{
  int length =(int) strlen(str);
  int i=offset;
  int size=0;
  while(i<length){
    if(str[i]==13 || str[i]=='\n'){
      return i+1;
    }
    if(i==speciali){
      *specialx=size;
    }
    size+=glutBitmapWidth(GLUT_BITMAP_HELVETICA_10,str[i]);
    
    if(size>maxSize){
      return i;
    }
    i++;
  }
  if(i==length && speciali==length){
    *specialx=size;
	}
  return i;
}


int getTextMaxSize(char * str,int offset,int maxSize)
{
  int length = strlen(str);
  int i=offset;
  int size=0;
  while(i<length){
    size+=glutBitmapWidth(GLUT_BITMAP_HELVETICA_10,str[i]);
    if(size>maxSize){
      return i;
    }
    i++;
  }
  return i;
}

int getCharWidth(char c)
{
  return glutBitmapWidth(GLUT_BITMAP_HELVETICA_10,c);
}

int getTextWidth(char * str)
{
  int size=0;
  if(!str)
    return 0;
  int i=0;
  int len=(int)strlen(str);
  while(i<len){
    size+=glutBitmapWidth(GLUT_BITMAP_HELVETICA_10,str[i++]);
  }
  return size;
}

/*
  post: returns 1 if the text went past the to value */
int drawTextFromTo(char * str, int from, int to)
{
  if(!str)
    return 0;
  int i=0;
  int size=0;
  int index=0;
  int len = strlen(str);
  /* Find the position in the text to start from */
  while(size<from && i<len){
    size+=glutBitmapWidth(GLUT_BITMAP_HELVETICA_10,str[i++]);
  }
  /* draw the text */
  while(i<len){
    size+=glutBitmapWidth(GLUT_BITMAP_HELVETICA_10,str[i]);
    if(size<to)
      glutBitmapCharacter(GLUT_BITMAP_HELVETICA_10 , str[i]);
    else break;
    i++;
  }
  if(size>=to)
    return 1;
  return 0;
}


GObject::GObject()
{
  init();
}
	
GObject::GObject(GObject * parent,char * text,int x,int y,int width,int height)
{
  init(parent,text,x,y,width,height);
}

void GObject::defaultAction(GObject * go)
{
}

void GObject::init()
{
  this->parent = parent;
  this->action=NULL;
  this->state=0;
  this->x=0;
  this->y=0;
  this->width=0;
  this->height=0;
  this->text=NULL;
  this->textWidth=0;
  setActive(1);
}

void GObject::init(GObject * parent,char * str,int x,int y,int width,int height)
{
  this->parent = parent;
  this->action=NULL;
  this->state=G_BORDER;
  this->x=x;
  this->y=y;
  this->width=width;
  this->height=height;
  this->text=NULL;
  this->textWidth=0;
  if(str){
    text = (char *)malloc(sizeof(char)*(strlen(str)+1));
    if(!text){
      //FIXME: error
    }
    sprintf(text,str);
    this->textWidth = getTextWidth(text);
  }
  setActive(1);
}

void GObject::getText(char * str,int maxlen)
{
  int lenToCopy;
  if(!text){
    str[0]=0;
    return;
  }
  lenToCopy=MIN(maxlen-1,(int)strlen(text));
  memcpy(str,text,lenToCopy);
  str[lenToCopy]=0;
}

void GObject::setText(char * str)
{
  if(!str){
    if(text){
      free(text);
      text=NULL;
    }
    textWidth=0;
    return;
  }
  if(!text){
    text = (char *)malloc(strlen(str)+1);
  }
  else if(strlen(str)>strlen(text)){
    free(text);
    text = (char *)malloc(strlen(str)+1);
  }
  
  if(!text){
    //FIXME: report error out of mem
  }
  textWidth = getTextWidth(text);
  sprintf(text,str); 
}

void GObject::release()
{
  if(text){
    free(text);
    text=NULL;
  }
}

GObject::~GObject()
{
  release();
}

void GObject::getPosition(int * x,int * y)
{
  *x=this->x;
  *y=this->y;
}
void GObject::getSize(int * w,int * h)
{
  *w=width;
  *h=height;
}

void GObject::setPosition(int x,int y)
{
  this->x=x;
  this->y=y;
}

int GObject::getYPosition()
{
  return y;
}
int GObject::getXPosition()
{
  return x;
}

void GObject::setXPosition(int x)
{
  this->x=x;
}

void GObject::setYPosition(int y)
{
  this->y=y;
}

int GObject::hit(int x,int y)
{
  if(x>=this->x && y>=this->y){
    if(x<(this->x+width) && y<(this->y+height))
      return 1;
  }
  return 0;
}

void GObject::drawHiShade()
{
  if(flipShade())
    shadeLoColor();
  else
    shadeHiColor();
  glBegin(GL_LINES);
   glVertex2i(x,y);
   glVertex2i(x+width,y);
   glVertex2i(x,y);
   glVertex2i(x,y+height);
  glEnd();
}

int GObject::flipShade()
{
  if(state&G_SELECTED)
    return 1;
  return 0;
}

void GObject::drawLoShade()
{
  if(flipShade())
    shadeHiColor();
  else
    shadeLoColor();
  glBegin(GL_LINES);
   glVertex2i(x+width,y);
   glVertex2i(x+width,y+height);
   glVertex2i(x,y+height);
   glVertex2i(x+width,y+height);
  glEnd();
}

void GObject::drawBackGround()
{
  bgColor();
  glBegin(GL_QUADS);
   glVertex2i(x,y);
   glVertex2i(x+width,y);
   glVertex2i(x+width,y+height);
   glVertex2i(x,y+height);
  glEnd();
  if(state&G_BORDER){
    drawHiShade();
    drawLoShade();
  }
}

void GObject::drawForeGround()
{
  fgColor();
  glRasterPos2i(x+2,y+TEXT_HEIGHT);
  drawText(text);
}

void GObject::draw()
{
  drawBackGround();
  drawForeGround();
}

int GObject::mousePressed(int button,int x,int y)
{
  if(hit(x,y) && getActive()){
    this->state|=G_SELECTED;
    return 1;
  }
  return 0;
}

int GObject::mouseReleased(int button,int x,int y)
{
  if(hit(x,y) && getActive()){
    this->state&=~G_SELECTED;
    if(action)
      action(this);
    else defaultAction(this);
    return 1;
  }
  this->state&=~G_SELECTED;
  return 0;
}

int GObject::mouseMoved(int x,int y)
{
  if(hit(x,y) && getActive()){
    this->state|=G_MOUSEOVER;
    return 1;
  }
  this->state&=~G_MOUSEOVER;
  return 0;
}

int GObject::mouseDragged(int x,int y)
{
  if(hit(x,y))
    return 1;
  else return 0;
}

void GObject::bgColor()
{
  glColor3f(0.5f,0.5f,0.5f);
}
void GObject::fgColor()
{
  glColor3f(0.0f,0.0f,0.0f);
}
void GObject::shadeHiColor()
{
  glColor3f(0.7f,0.7f,0.7f);
}
void GObject::shadeLoColor()
{
  glColor3f(0.3f,0.3f,0.3f);
}




