#ifndef G_OBJECT_H
#define G_OBJECT_H

#define G_MOUSEOVER 0x1
#define G_SELECTED  0x2
#define G_SHOWING    0x4
#define G_BORDER    0x8
#define G_ACTIVE   0x10
#define G_GETWIDTHFROMTEXT    0x20
#define G_DESTROY 0x40
#include<GL/gl.h>
#include <GL/glut.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#define MAX(a,b) (((a)>(b))?(a):(b))
#define MIN(a,b) (((a)<(b))?(a):(b))

#ifdef _WIN32
#define DIR_SEPARATOR '\\' 
#else
#define DIR_SEPARATOR '/'
#endif

int  getCharWidth(char c);
int  getTextWidth(char * str);
int drawTextFromTo(char * str, int from, int to);
void drawText(const char *,...);
int  getTextMaxSize(char * str,int offset,int maxSize);
int  getTextMaxSizeEx(char * str,int offset,int maxSize,int speciali,int * specialx);

typedef struct GColor_T
{
	float r,g,b;
}GColor;

#define TEXT_HEIGHT 10

class GObject
{
protected:
	GObject * parent;
	int x,y;
	int width,height;
	char * text;
	int textWidth;
	unsigned int id;
public:
	int state;
	void (* action)(GObject *);

	GObject();
	GObject(GObject * parent,char * text,int x,int y,int width,int height);
	virtual ~GObject();

	virtual void bgColor();
	void fgColor();
	void shadeHiColor();
	void shadeLoColor();

	virtual int flipShade();
	void drawHiShade();
	void drawLoShade();

	void drawBackGround();
	void drawForeGround();

	virtual void release();
	virtual void init();
	virtual void init(GObject * parent,char * str,int x,int y,int width,int height);
	
	int getYPosition();
	int getXPosition();
	int getHeight(){return height;}
	int getWidth(){return width;}


	GObject * getParent(){return parent;};
	void getPosition(int * x,int * y);
	void getSize(int * w,int * h);

	virtual void setText(char * t);
	void getText(char * str,int maxlen);

	int awaitingDestruction(){return state&G_DESTROY;}
	void postDestroy() {state|=G_DESTROY;}
	virtual void setShowing(int value) {if(value) state|=G_SHOWING; else state&=~G_SHOWING;}
	virtual int isShowing(){return state&G_SHOWING;};
	virtual int getShowing(){return isShowing();};

	virtual void setActive(int value) {if(value) state|=G_ACTIVE; else state&=~G_ACTIVE;}
	virtual int getActive() {return state&G_ACTIVE;}
	virtual int getIsSelected() {return state&G_SELECTED;}
	virtual void setIsSelected(int i){
	  if(i)
	    state|=G_SELECTED;
	  else state&=~G_SELECTED;}


	void setWidth(int w){width=w;}
	void setHeight(int h){height=h;}
	void setParent(GObject * p){parent=p;}
	virtual void setPosition(int x,int y);
	virtual void setXPosition(int x);
	virtual void setYPosition(int y);
	virtual int hit(int x,int y);
	virtual void draw();

	virtual int keyTyped(unsigned int mod,char key) {return 0;};
	virtual int mousePressed(int button,int x,int y);
	virtual int mouseReleased(int button,int x,int y);
	virtual int mouseMoved(int x,int y);
	virtual int mouseDragged(int x,int y);
	virtual void defaultAction(GObject *);
};

class GStatic : public GObject
{
public:
  GStatic(GObject * parent,char * str,int x,int y,int width,int height)
    {init(parent,str,x,y,width,height);state=0;}
  virtual void draw() {drawBackGround();drawForeGround();}
};

class GGroupBox : public GObject
{
 public:
  GGroupBox(GObject * parent,char * str,int x,int y,int width,int height)
    {
      init(parent,str,x,y,width,height);
    }
  virtual void draw() {
    shadeLoColor();
    glBegin(GL_LINE_LOOP);
     glVertex2i(x,y);
     glVertex2i(x+width,y);
     glVertex2i(x+width,y+height);
     glVertex2i(x,y+height);
    glEnd();
		
    bgColor();
    glBegin(GL_QUADS);
     glVertex2i(x+2,y-TEXT_HEIGHT/2);
     glVertex2i(x+2+textWidth,y-TEXT_HEIGHT/2);
     glVertex2i(x+2+textWidth,y+TEXT_HEIGHT/2);
     glVertex2i(x+2,y+TEXT_HEIGHT/2);
    glEnd();
    fgColor();
    glRasterPos2i(x+2,y+TEXT_HEIGHT/2-2);
    drawText(text);
  }
  virtual int mousePressed(int button,int x,int y) {return 0;}
  virtual int mouseReleased(int button,int x,int y) {return 0;}
  virtual int mouseMoved(int x,int y){return 0;}
  virtual int mouseDragged(int x,int y){return 0;}
};

#endif


