#include "GScrollBar.h"
#include "GScrollable.h"


void b1Action(GObject * go)
{
  if(go && go->getParent() && (go->getParent())->getParent()){
    GScrollable * gs = (GScrollable * )go->getParent()->getParent();
    gs->scrollAction1(go->getParent());//send the scrollbar
  }
}

void b2Action(GObject * go)
{
  if(go && go->getParent() && go->getParent()->getParent()){
    GScrollable * gs = (GScrollable * )go->getParent()->getParent();
    gs->scrollAction2(go->getParent());//send the scrollbar
  }
}

GScrollBar::GScrollBar()
{
  init();
  state|=G_VERTICAL_SCROLL;
  initButtons();
}

GScrollBar::GScrollBar(int flags, GObject * parent, int x,int y,int width,int height)
{
  init(parent,"",x,y,width,height);
  state|=flags;
  initButtons();
}

void GScrollBar::initButtons()
{
  if(state&G_VERTICAL_SCROLL){
    b1=new GButton(this,"^",x,y,width,width);
    b2=new GButton(this,"v",x,y+height-width,width,width);
    b1->action = b1Action;
    b2->action = b2Action;
  }
  else{
    b1 = new GButton(this,"<",x,y,height,height);
    b2 = new GButton(this,">",x+width-height,y,height,height);
    b1->action = b1Action;
    b2->action = b2Action;
  }
}

void GScrollBar::release()
{
  if(b1){
    delete b1;
    b1=NULL;
  }
  if(b2){
    delete b2;
    b2=NULL;
  }
}

GScrollBar::~GScrollBar()
{
  release();
}

void GScrollBar::setHeight(int h)
{
  GObject::setHeight(h);
  if(state & G_VERTICAL_SCROLL){
    b2->setYPosition(y+h-b2->getHeight());
  }
  else{
    b1->setHeight(h);
    b2->setHeight(h);
  }
}

void GScrollBar::setWidth(int w)
{
  GObject::setWidth(w);
  if(state & G_HORIZONTAL_SCROLL){
    b2->setXPosition(x+w-b2->getWidth());
  }
  else{
    b1->setWidth(w);
    b2->setWidth(w);
  }
}

void GScrollBar::setXPosition(int x)
{
  int offset = x-this->x;
  GObject::setXPosition(x);
  if(b1){
    b1->setXPosition(b1->getXPosition()+offset);
  }
  if(b2){
    b2->setXPosition(b2->getXPosition()+offset);
  }
}
void GScrollBar::draw()
{
  float ratio;
  int bheight;
  int bx;
  int bwidth;
  int by;

  if(state&G_VERTICAL_SCROLL){
    float totalHeight = (float) ((GScrollable*)parent)
      ->getScrollAreaSize(this);
    float top = ( float )((GScrollable *)parent)
      ->getScrollPosition(this);
    ratio = ((float)(height))/totalHeight;
    bx=x;
    bwidth=width;
    by=y+(int)(top*((float)(height-(2*b1->getHeight()+2)))/totalHeight)
      +b1->getHeight();
    if(ratio>=1.0f)
      ratio=1.0f;
    bheight=(int)((height-(2*b1->getHeight())+2)*ratio);
    if(bheight<b1->getHeight())
      bheight=b1->getHeight();
  }
  else {
    float totalWidth = (float) ((GScrollable *)parent)
      ->getScrollAreaSize(this);
    float left = (float)((GScrollable *)parent)
      ->getScrollPosition(this);
    ratio = ((float)(width))/totalWidth;
    bx=x+(int)(left*((float)(width-(2*b1->getWidth()+2)))/totalWidth)
      +b1->getWidth();
    if(ratio>=1.0f)
      ratio=1.0f;
    bwidth=(int)((width-(2*b1->getWidth()+2))*ratio);
    by=y;
    bheight=height;
    if(bwidth<b1->getWidth())
      bwidth=b1->getWidth();
  }
  if(state&G_VERTICAL_SCROLL && ratio>=1.0f)
    return;
  shadeHiColor();
  glBegin(GL_QUADS);
  glVertex2i(x,y);
  glVertex2i(x,y+height);
  glVertex2i(x+width,y+height);
  glVertex2i(x+width,y);
  glEnd();
  if(b1)
    b1->draw();
  if(b2)
    b2->draw();
  bgColor();
  glBegin(GL_QUADS);
   glVertex2i(bx,by);
   glVertex2i(bx,by+bheight);
   glVertex2i(bx+bwidth,by+bheight);
   glVertex2i(bx+bwidth,by);
  glEnd();
  glBegin(GL_LINES);
  shadeHiColor();
   glVertex2i(bx,by);
   glVertex2i(bx+bwidth,by);
   glVertex2i(bx,by);
   glVertex2i(bx,by+bheight);
  shadeLoColor();
   glVertex2i(bx,by+bheight);
   glVertex2i(bx+bwidth,by+bheight);
   glVertex2i(bx+bwidth,by);
   glVertex2i(bx+bwidth,by+bheight);
  glEnd();
  
}

void GScrollBar::scrollBarPressed(int x,int y)
{
  if(state&G_VERTICAL_SCROLL){
    //Send the paramaters as the numerator and denominator of the ratio
    ((GScrollable *)parent)->scrollBarPressed(this,
					      this->height-(2*b1->getHeight()),y-this->y-b1->getHeight());
  }
  else {
     ((GScrollable *)parent)->scrollBarPressed(this,
					      this->width-(2*b1->getWidth()),x-this->x-b1->getWidth());
    
  }
}

int GScrollBar::mousePressed(int button,int x,int y)
{
  if(hit(x,y)){
    int ret=b1->mousePressed(button,x,y);
    ret|=b2->mousePressed(button,x,y);
    if(!ret){
      scrollBarPressed(x,y);
    }
    return 1;
  }
  return 0;
}
int GScrollBar::mouseReleased(int button,int x,int y)
{
  int ret=b1->mouseReleased(button,x,y);
  ret|=b2->mouseReleased(button,x,y);
  return ret;
}

int GScrollBar::mouseDragged(int x,int y)
{
  //FIXME:
  if(hit(x,y)){
    if(!b1->hit(x,y) && !b2->hit(x,y)){
      scrollBarPressed(x,y);
      return 1;
    }

  }
  return 0;
}









