#include "GWindow.h"
#include "GButton.h"
#include "GWindows.h"

void drawCloseButton(GObject * go)
{
  int x,y;
  int w,h;
  if(go){
    go->getPosition(&x,&y);
    go->getSize(&w,&h);
    go->drawBackGround();
    go->fgColor();
    
    glBegin(GL_LINES);
    glVertex2i(x+1,y+1);
    glVertex2i(x+w-2,y+h-2);
    glVertex2i(x+1,y+h-2);
    glVertex2i(x+w-2,y+1);
    glEnd();
  }
}

void closeButtonActionDestroy(GObject * go)
{
  if(go && go->getParent()){
    go->getParent()->postDestroy();
  }
}

void closeButtonActionDisactivate(GObject * go)
{
  if(go && go->getParent()){
    go->getParent()->setShowing(0);
    go->state&=~G_SELECTED;
  }
}

GWindow::GWindow()
{
  resizing=moving=0;
  init();
  
  closeButton=NULL;
  initObjects();
}

GWindow::GWindow(GObject * parent,char * str,int x,int y,int width,int height)
{
  resizing=moving=0;
  init(parent,str,x,y,width,height);
  initObjects();
  
  state|=G_TITLEBAR|G_MOVEABLE|G_RESIZEABLE;
  minWidth=TEXT_HEIGHT*(strlen(str)+1);
  minHeight=TEXT_HEIGHT+8;
  
  closeButton=NULL;
  if(G_TITLEBAR){
    closeButton = new GButton(this,"x",width-TEXT_HEIGHT-4,4,TEXT_HEIGHT,TEXT_HEIGHT);
    closeButton->drawFunc=drawCloseButton;
    closeButton->action = closeButtonActionDisactivate;
    addObject(closeButton);
  } 
}

void GWindow::destroyOnClose()
{
  if(closeButton){
    closeButton->action = closeButtonActionDestroy;
  }
}

void GWindow::initObjects()
{
  numObjects=0;
  maxObjects=20;
  objects = (GObject **) malloc(sizeof(GObject *)*maxObjects);
  if(!objects)
    ;//FIXME
  memset(objects,0,sizeof(GObject *)*maxObjects);
}

void GWindow::release()
{
  int i=0;
  if(objects){
    while(i<numObjects){
      if(objects[i]){
	delete objects[i];
      }
      i++;
    }
    free(objects);
  }
}

GWindow::~GWindow()
{
  release();
}

int GWindow::flipShade()
{
  return 0;
}

void GWindow::drawTitleBar()
{
  titleBarColor();
  glBegin(GL_QUADS);
   glVertex2i(x+2,y+2);
   glVertex2i(x+width-2,y+2);
   glVertex2i(x+width-2,y+TEXT_HEIGHT+4);
   glVertex2i(x+2,y+TEXT_HEIGHT+4);
  glEnd();

  fgColor();
  glRasterPos2i(x+4,y+2+TEXT_HEIGHT);
  drawText(this->text);
}

void GWindow::draw()
{
  drawBackGround();
  if(state&G_TITLEBAR){
    drawTitleBar();
  }
  
  glPushMatrix();
  glLoadIdentity();
  glTranslatef(x,y,0);
  int i=numObjects-1;
  while(i>=0){
    objects[i]->draw();
    i--;
  }
  glPopMatrix();
}

void GWindow::setShowing(int i)
{
  GObject::setShowing(i);
  if(i)
    gwIncreasePrecedencePointer(this);
  else{
  }  
}

void GWindow::addObject(GObject * go)
{
  if(numObjects<maxObjects){
    //Ignore the close button since it moves with the window
    if(go!=closeButton){
      int x,y,w,h;
      go->getPosition(&x,&y);
      go->getSize(&w,&h);
      if(x+w > minWidth){
	minWidth = x+w;
      }
      if(y+h > minHeight)
	minHeight = y+h;
    }
    objects[numObjects]=go;
    numObjects++;
  }
  //else FIXME
}

int GWindow::keyTyped(unsigned int mod,char key)
{
  int i=0;
  while(i<numObjects){
    if(objects[i]->keyTyped(mod,key))
      return 1;
    i++;
  }
  return 0;
}

void GWindow::increasePrecedence(int index)
{
  if(index!=0){
    GObject * back = objects[0];
    objects[0]=objects[index];
    objects[index]=back;
  }
}

int GWindow::mousePressed(int button,int x,int y)
{
  int i=0;
  int ret=0;
  
  if(hit(x,y)){
    state|=G_SELECTED;
    
    while(i<numObjects){
      if(objects[i]->mousePressed(button,x-this->x,y-this->y)){
	ret=1;
	increasePrecedence(i);
	break;
      }
      else if(objects[i]->getIsSelected()){
	objects[i]->setIsSelected(0);
      }
      i++;
    }
    i++;
    while(i<numObjects){
      if(objects[i]->getIsSelected()){
	objects[i]->setIsSelected(0);
      }
      i++;
    }

    if(state&G_TITLEBAR){
      if(y<(this->y+TEXT_HEIGHT+6)){
	if(state&G_MOVEABLE){
	  oldx=x;
	  oldy=y;
	  moving=1;		
	}
	return 1;
      }
    }
    if(state&G_RESIZEABLE){
      resizing=0;
      oldx=oldy=-1;
      if((this->x+width-x)<=3){
	resizing=1;
	oldx=1;
      }
      if((this->y+height-y)<=3){
	resizing=1;
	oldy=1;
      }
    }
    return 1;
  }
  else state&=~G_SELECTED;
  return ret;
}
int GWindow::mouseReleased(int button,int x,int y)
{
  int i=0;
  int ret=0;
  
  moving=0;
  resizing=0;
  
  while(i<numObjects){
    ret|=objects[i]->mouseReleased(button,x-this->x,y-this->y);
    i++;
  }
  return ret;
}
int GWindow::mouseMoved(int x,int y)
{
  int i=0;
  int ret =0;
  while(i<numObjects){
    ret|=objects[i]->mouseMoved(x-this->x,y-this->y);
    i++;
  }
  return ret;
}

int GWindow::mouseDragged(int x,int y)
{
  int i=0;
  int ret=0;
  if(hit(x,y)){
    while(i<numObjects){
      if(objects[i]->mouseDragged(x-this->x,y-this->y)){
	ret=1;
	break;
      }
      i++;
    }
    if((state&G_MOVEABLE) && moving){   
      this->x+=(x-oldx);
      this->y+=(y-oldy);
      oldx=x;
      oldy=y;
      ret=1;
    }
    else if((state&G_RESIZEABLE) && resizing){
      if(oldx>0){
	
	this->width = x-this->x;
	if(width<minWidth)
	  width=minWidth;
	if(closeButton)
	  closeButton->setXPosition(width-TEXT_HEIGHT-4);
      }
      if(oldy>0){
	this->height= y-this->y;
	if(height<minHeight)
	  height=minHeight;
      }
      ret=1;
    }
    return 1;
  }
  return ret;
}


void GWindow::titleBarColor()
{
  glColor3f(0.0f,0.0f,0.7f);
  if(state&G_SELECTED)
    glColor3f(0.2f,0.2f,0.8f);
}








