#include"useful.h"
#include<stdlib.h>
#include<string.h>
#include<math.h>
#include<GL/gl.h>
#include<GL/glext.h>
#include"mesh.h"
#include "pca.h"

float dot(float a[3],float b[3])
{
  return (a[0]*b[0] + a[1]*b[1] + a[2]*b[2]);
}

void cross(float a[3],float b[3],float c[3])
{
  a[0] = b[1]*c[2] - b[2]*c[1];
  a[1] = b[2]*c[0] - b[0]*c[2];
  a[2] = b[0]*c[1] - b[1]*c[0];
}

float magnitude(float a[3])
{
  return(float)sqrt(a[0]*a[0]+a[1]*a[1]+a[2]*a[2]);
}

void unify(float a[3])
{
  float len=magnitude(a);
  a[0]/=len;
  a[1]/=len;
  a[2]/=len;
}

Mesh::Mesh()
{
  tex=0;
  list=0;
  pos[0]=pos[1]=pos[2]=0;
  eul[0]=eul[1]=eul[2]=0;
  nvert=0;
  vert=0;
  vnorm=0;
  npoly=0;
  poly=0;
  pnorm=0;

  memset(rotAxis,0,sizeof(float)*9);
  rotAxis[2]=1;
  rotAxis[4]=-1;
  rotAxis[6]=1;
  /* HAND
  rotAxis[2]=1;
  rotAxis[4]=1;
  rotAxis[8]=1;
  */
}

Mesh::~Mesh()
{
  destroy();
}

void Mesh::destroy()
{
  if(nvert)
  {
    delete vert;
    delete vnorm;
    nvert=0;
  }
  if(npoly)
  {
    delete poly;
    delete pnorm;
    npoly=0;
  }
  if(tex)glDeleteTextures(1,&tex);
  if(list)glDeleteLists(list,1);
}

void Mesh::setRotAxis(float * a)
{
  memcpy(rotAxis,a,sizeof(float)*9);
}


int Mesh::loadobj_gz(gzFile gz)
{
  int i=0;
  char buf[256];

  gzGetLine(gz,buf,255);
  sscanf(buf,"%d",&nvert);
  printDebug("in Mesh::loadobj_gz\n");
  printDebug("nvert is %d\n",nvert);
  vert = new _3floats[nvert];
  vnorm=new float[nvert][3];
  i=0;
  while(i<nvert){
    gzGetLine(gz,buf,255);
    sscanf(buf,"%f %f %f",vert[i],vert[i]+1,vert[i]+2);
    i++;
  }
  gzGetLine(gz,buf,255);
  sscanf(buf,"%d",&ntvert);
  printDebug("ntvert is %d\n",ntvert);
  tvert = new _2floats[ntvert];
  for(i=0;i<ntvert;i++)
  {
    gzGetLine(gz,buf,255);
    sscanf(buf,"%f %f",tvert[i],tvert[i]+1);
  }
  gzGetLine(gz,buf,255);
  sscanf(buf,"%d",&npoly);
  printDebug("npoly is %d\n",npoly);
  tpoly = new _3ints[npoly];
  poly = new _3ints[npoly];
  pnorm=new float[npoly][3];
  for(i=0;i<npoly;i++)
  {
    gzGetLine(gz,buf,255);
    sscanf(buf,"%d %d %d",poly[i],poly[i]+1,poly[i]+2);
    tpoly[i][0]=poly[i][0];
    tpoly[i][1]=poly[i][1];
    tpoly[i][2]=poly[i][2];
  }
  compute_normals();
  createlist();
  printDebug("returning from Mesh::loadobj_gz\n");
  return 1;
}

int Mesh::loadobj(char *fname)
{
  FILE *f;
  int r=loadobj(f=fopen(fname,"rb"));
  fclose(f);
  return r;
}

int Mesh::loadobj(FILE *fd)
{
  char *line;
  int i;
  if(fd==NULL) return 0;

  nvert=linecount(fd,"v ");
  ntvert=linecount(fd,"vt ");
  npoly=linecount(fd,"f ");

  vert=new _3floats[nvert];
  tvert=new _2floats[ntvert];
  vnorm=new float[nvert][3];
  poly=new _3ints[npoly];
  tpoly=new _3ints[npoly];
  pnorm=new float[npoly][3];

  for(i=0;i<nvert;i++)
  {
    line=nextline(fd,"v ");
    sscanf(line,"v %f %f %f",vert[i],vert[i]+1,vert[i]+2);
  }
  
  fseek(fd,0,SEEK_SET);
  
  for(i=0;i<ntvert;i++)
  {
    line=nextline(fd,"vt ");
    sscanf(line,"vt %f %f",tvert[i],tvert[i]+1);
  }

  fseek(fd,0,SEEK_SET);

  for(i=0;i<npoly;i++)
  {
    line=nextline(fd,"f ");
    sscanf(line,"f %d %d %d",poly[i],poly[i]+1,poly[i]+2);
    poly[i][0]--;
    poly[i][1]--;
    poly[i][2]--;
    tpoly[i][0]=poly[i][0];
    tpoly[i][1]=poly[i][1];
    tpoly[i][2]=poly[i][2];
  }

  compute_normals();
  createlist();
  return 1;
}

int Mesh::loadlwo(char *fname)
{
  FILE *f;
  int r=loadlwo(f=fopen(fname,"rb"));
  fclose(f);
  return r;
}

int Mesh::loadlwo(FILE *fd)
{
  int i,j,l,n=0;
  int file_length,cur_pos=0,poly_pos,vert_pos;
  unsigned short int s;
  float f;
  int temp;

  if(fd==NULL) return 0;

  destroy();

  // check header
  fread(&l,4,1,fd);
  cur_pos+=4;

  if(swaplong(l)!=FORM)
  {
    fprintf(stderr,"not an IFF file!\n");
    return 0;
  }
  fread(&l,4,1,fd);
  cur_pos+=4;
  file_length=swaplong(l);

  fread(&l,4,1,fd);
  cur_pos+=4;

  if(swaplong(l)!=LWOB)
  {
    fprintf(stderr,"not a lightwave object!\n");
    return 0;
  }

  // find data
  while(1)
  {
    if(cur_pos>file_length || feof(fd))
    {
      fprintf(stderr,"file length error!\n");
      return 0;
    }

    fread(&l,4,1,fd);
    cur_pos+=4;

    l=swaplong(l);

    if (l==PNTS) 
    {
      vert_pos=cur_pos;
    }
    else if (l==POLS)
    {
      poly_pos=cur_pos;
      break;
    }

    fread(&n,4,1,fd);
    cur_pos+=4;
    n=swaplong(n);

    fseek(fd,n,SEEK_CUR);
    cur_pos+=n;
    if(n==0)
    {
      fprintf(stderr,"bad chunk length!");
      return 0;
    }

  }

//------ read vertex data

  fseek(fd,vert_pos,SEEK_SET);

  fread(&n,4,1,fd);
  n=swaplong(n);

  nvert=n/12;
  vert=new float[nvert][3];
  vnorm=new float[nvert][3];

  for(i=0;i<nvert;i++)
  {
    fread(&f,4,1,fd);
    vert[i][0]=swapfloat(f);
    fread(&f,4,1,fd);
    vert[i][1]=-swapfloat(f);
    fread(&f,4,1,fd);
    vert[i][2]=-swapfloat(f);
  }

//------ read polygon data

  fseek(fd,poly_pos,SEEK_SET);

  fread(&n,4,1,fd);
  n=swaplong(n);

  npoly=n/10;
  poly=new int[npoly][3];
  pnorm=new float[npoly][3];

  for(i=0;i<npoly;i++)
  {
    fread(&s,2,1,fd);
    s=swapshort(s);

    if(s!=3)
    {
      fprintf(stderr,"error: non-triangular polygon found!(polygon number %d is %d verts)\n",i,n);
      exit(-1);
    }
    for(j=0;j<3;j++)
    {
      fread(&s,2,1,fd);
      poly[i][j]=swapshort(s);
    }
    temp=poly[i][0];
    poly[i][0]=poly[i][1];
    poly[i][1]=temp;
    fseek(fd,2,SEEK_CUR);
  }

  compute_normals();
  createlist();

  return 1;
}

void Mesh::maketexturecoords(int dim[2],float scale[2])
{
  float min[2],max[2],range[2];
  int i,j;

  for(j=0;j<2;j++)
  {
    min[j]=scale[j]*vert[0][dim[j]];
    max[j]=scale[j]*vert[0][dim[j]];
  }

  for(i=1;i<nvert;i++)
  {
    for(j=0;j<2;j++)
    {
     if(scale[j]*vert[i][dim[j]]<min[j])min[j]=scale[j]*vert[i][dim[j]];
     if(scale[j]*vert[i][dim[j]]>max[j])max[j]=scale[j]*vert[i][dim[j]];
    }
  }
  for(j=0;j<2;j++) range[j]=max[j]-min[j];

  ntvert=nvert;
  tvert=new _2floats[ntvert];

  for(i=1;i<nvert;i++)
  {
    tvert[i][0]=(scale[0]*vert[i][dim[0]]-min[0])/range[0];
    tvert[i][1]=(scale[1]*vert[i][dim[1]]-min[1])/range[1];
  }
  createlist();
}

void Mesh::compute_normals()
{
  int i,j;

  compute_pnormals();

  // use average of poly normals for vert normals

  for(i=0;i<nvert;i++)
  {
    vnorm[i][0]=vnorm[i][1]=vnorm[i][2]=0;
    for(j=0;j<npoly;j++)
    {
      if(poly[j][0]==i || poly[j][1]==i || poly[j][2]==i) 
      {
        vnorm[i][0] += pnorm[j][0];
        vnorm[i][1] += pnorm[j][1];
        vnorm[i][2] += pnorm[j][2];
      }
    }
    unify(vnorm[i]);
  }
}

void Mesh::compute_pnormals()
{
  int i,j;
  float vec1[3],vec2[3];

  for(i=0;i<npoly;i++)
  {
    // vec1 = point1 - point0
    // vec2 = point2 - point1
    for(j=0;j<3;j++)
    {
      vec1[j]=vert[poly[i][1]][j] - vert[poly[i][0]][j];
      vec2[j]=vert[poly[i][1]][j] - vert[poly[i][2]][j];
    }
    // polygon normal = vec1 X vec2
    cross(pnorm[i],vec1,vec2);
    unify(pnorm[i]);
  }
}

void Mesh::createlist()
{
  if(!glIsList(list))list=glGenLists(1);
  glNewList(list,GL_COMPILE);
  
  glBegin(GL_TRIANGLES);

  for(int i=0;i<npoly;i++)
  {
    if(vnorm)glNormal3fv(vnorm[poly[i][0]]);
    if(tvert)
    {
      glTexCoord2fv(tvert[tpoly[i][0]]);
      if(Pca::getHardwareSupported()){
	//glMultiTexCoord2fvARB(GL_TEXTURE0_ARB,tvert[tpoly[i][0]]);
	glMultiTexCoord2fvARB(GL_TEXTURE1_ARB,tvert[tpoly[i][0]]);
	glMultiTexCoord2fvARB(GL_TEXTURE2_ARB,tvert[tpoly[i][0]]);
	glMultiTexCoord2fvARB(GL_TEXTURE3_ARB,tvert[tpoly[i][0]]);
      }
    }
    glVertex3fv(vert[poly[i][0]]);
    if(vnorm)glNormal3fv(vnorm[poly[i][1]]);
    if(tvert)
    {
      glTexCoord2fv(tvert[tpoly[i][1]]);
      if(Pca::getHardwareSupported()){
	//glMultiTexCoord2fvARB(GL_TEXTURE0_ARB,tvert[tpoly[i][1]]);
	glMultiTexCoord2fvARB(GL_TEXTURE1_ARB,tvert[tpoly[i][1]]);
	glMultiTexCoord2fvARB(GL_TEXTURE2_ARB,tvert[tpoly[i][1]]);
	glMultiTexCoord2fvARB(GL_TEXTURE3_ARB,tvert[tpoly[i][1]]);
      }
    }
    glVertex3fv(vert[poly[i][1]]);
    if(vnorm)glNormal3fv(vnorm[poly[i][2]]);
    if(tvert)
    {
      glTexCoord2fv(tvert[tpoly[i][2]]);
      if(Pca::getHardwareSupported()){
	//glMultiTexCoord2fvARB(GL_TEXTURE0_ARB,tvert[tpoly[i][2]]);
	glMultiTexCoord2fvARB(GL_TEXTURE1_ARB,tvert[tpoly[i][2]]);
	glMultiTexCoord2fvARB(GL_TEXTURE2_ARB,tvert[tpoly[i][2]]);
	glMultiTexCoord2fvARB(GL_TEXTURE3_ARB,tvert[tpoly[i][2]]);
      }
    }
    glVertex3fv(vert[poly[i][2]]);
  }

  glEnd();
  glEndList();
}

void Mesh::draw()
{
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
    glTranslatef(pos[0],pos[1],pos[2]);
    glRotatef(eul[0],rotAxis[0],rotAxis[1],rotAxis[2]);//0,0,1);
    // glRotatef(eul[1],rotAxis[3],rotAxis[4],rotAxis[5]);//0,-1,0);
    glRotatef(eul[2],rotAxis[6],rotAxis[7],rotAxis[8]);//1,0,0);
    glRotatef(eul[1],rotAxis[3],rotAxis[4],rotAxis[5]);//0,-1,0);
    glCallList(list);
  glPopMatrix();
}
