#include<GL/gl.h>
#include <GL/glut.h>
#include<stdlib.h>
#include<stdio.h>
#include <ctype.h>
#include"pca.h"
#include"useful.h"


int Pca::hardwareSupported=0;
int Pca::numCreated=0;

int Pca::getHardwareSupported()
{
  static int called=0;
  if(!called){
    char  buffer[1024];
    
    int i=0;
    Pca::hardwareSupported=0;

    sprintf(buffer,"%s",glGetString(GL_VENDOR));
    while(i<strlen(buffer)){
      buffer[i]=tolower(buffer[i]);
      i++;
    }
    if(strstr(buffer,"nvidia")){
      sprintf(buffer,"%s",glGetString(GL_RENDERER));
      i=0;
      while(i<strlen(buffer)){
	buffer[i]=tolower(buffer[i]);
	i++;
      }
      if(strstr(buffer,"geforce")){
	Pca::hardwareSupported=1;
      }
    }
    called=1;
  }
  return Pca::hardwareSupported;
}

Pca::Pca(){
  
  pcaRenderer = new PcaRendererSoft(3);
  nrchannels=3;

  Pca::numCreated++;
}
Pca::Pca(int num_channels){
  
  pcaRenderer = new PcaRendererSoft(num_channels);
  nrchannels=num_channels;

  Pca::numCreated++;
}
void Pca::destroy()
{
  if(pcaRenderer)
    delete pcaRenderer;
  pcaRenderer=NULL;
}
void PcaRenderer::initNumChannels(int num_channels)
{
//  static float YUVtoRGB[12]={1  , 1  , 1, 1,
//                             0     ,-0.813, 2.018 , 1,
//		             2.018,-0.391 , 0     , 1};
  static float YUVtoRGB[12]={1 , 1 , 1 , 1,
                             0 , 0 , 1 , 1,
		             1 , 0 , 0 , 1};
/*  static float YUVtoRGB[12]={1  , 0  , 0, 1,
                             0     ,1, 0 , 1,
		             0,0 , 1     , 1};*/
  nrchannels=num_channels;
  nreig=new int[nrchannels];
  nruse=new int[nrchannels];
  eigentex=new GLuint*[nrchannels];
  colormatrix=new float[nrchannels*4];
  
  memset(nreig,0,sizeof(int)*nrchannels);
  memset(nruse,0,sizeof(int)*nrchannels);
  memset(eigentex,0,sizeof(GLuint *)*nrchannels);

  if(nrchannels==3) {// default to YUV color conversion
    for(int i=0;i<nrchannels*4;i++)
      colormatrix[i]=YUVtoRGB[i];
     // colormatrix[i]=(YUVtoRGB[i]+1)/2;
  }
  else {
    for(int i=0;i<nrchannels*4;i++)
      colormatrix[i]=1;
  }

  printDebug("done initing num channels\n");
}

PcaRendererNV::PcaRendererNV(int num_channels)
{
  int ntexunits;
  initNumChannels(num_channels);

  glGetIntegerv(GL_MAX_TEXTURE_UNITS_ARB,&ntexunits);
  if(ntexunits<2)
  {
    printf("your graphics card is not good enuf. goodbye\n");
    exit(0);
  }
  else if(ntexunits>=4)
  {
    glActiveTextureARB(GL_TEXTURE0_ARB);
    glEnable(GL_TEXTURE_2D);
    glActiveTextureARB(GL_TEXTURE1_ARB);
    glEnable(GL_TEXTURE_2D);
    glActiveTextureARB(GL_TEXTURE2_ARB);
    glEnable(GL_TEXTURE_2D);
    glActiveTextureARB(GL_TEXTURE3_ARB);
    glEnable(GL_TEXTURE_2D);

    glEnable(GL_PER_STAGE_CONSTANTS_NV);

    setup_pca_state=&PcaRendererNV::setup_pca_state_NV20;
    _renderpca=&PcaRendererNV::renderpca_NV20;
  }
  else
  {
    glActiveTextureARB(GL_TEXTURE0_ARB);
    glEnable(GL_TEXTURE_2D);
    glActiveTextureARB(GL_TEXTURE1_ARB);
    glEnable(GL_TEXTURE_2D);
/*    glActiveTextureARB(GL_TEXTURE2_ARB);
    glDisable(GL_TEXTURE_2D);
    glActiveTextureARB(GL_TEXTURE3_ARB);
    glDisable(GL_TEXTURE_2D);

    glDisable(GL_PER_STAGE_CONSTANTS_NV);*/

    setup_pca_state=&PcaRendererNV::setup_pca_state_NV17;
    _renderpca=&PcaRendererNV::renderpca_NV17;
  }
  
  pcalist=glGenLists(1);

  glNewList(pcalist,GL_COMPILE);
  (this->*setup_pca_state)();
  glBegin(GL_QUADS);
   glMultiTexCoord2fARB(GL_TEXTURE0_ARB,0,0);
   glMultiTexCoord2fARB(GL_TEXTURE1_ARB,0,0);
   if(ntexunits>=4)
   {
     glMultiTexCoord2fARB(GL_TEXTURE2_ARB,0,0);
     glMultiTexCoord2fARB(GL_TEXTURE3_ARB,0,0);
   }
   glVertex2f(0,0);

   glMultiTexCoord2fARB(GL_TEXTURE0_ARB,0,1);
   glMultiTexCoord2fARB(GL_TEXTURE1_ARB,0,1);
   glMultiTexCoord2fARB(GL_TEXTURE2_ARB,0,1);
   glMultiTexCoord2fARB(GL_TEXTURE3_ARB,0,1);
   glVertex2f(0,1);
   
   glMultiTexCoord2fARB(GL_TEXTURE0_ARB,1,1);
   glMultiTexCoord2fARB(GL_TEXTURE1_ARB,1,1);
   glMultiTexCoord2fARB(GL_TEXTURE2_ARB,1,1);
   glMultiTexCoord2fARB(GL_TEXTURE3_ARB,1,1);
   glVertex2f(1,1);

   glMultiTexCoord2fARB(GL_TEXTURE0_ARB,1,0);
   glMultiTexCoord2fARB(GL_TEXTURE1_ARB,1,0);
   glMultiTexCoord2fARB(GL_TEXTURE2_ARB,1,0);
   glMultiTexCoord2fARB(GL_TEXTURE3_ARB,1,0);
   glVertex2f(1,0);
  glEnd();
  glEndList();
  setReduceFactor(1);

  basisWidth = new int[nrchannels];
  memset(basisWidth,0,sizeof(int)*nrchannels);
  basisHeight = new int[nrchannels];
  memset(basisHeight,0,sizeof(int)*nrchannels);
}

PcaRendererNV::~PcaRendererNV()
{
  destroy();
}

void PcaRendererNV::destroy()
{
  int i=0;
  glDeleteLists(pcalist,1);
  while(i<nrchannels){
    if(eigentex[i]){
      glDeleteTextures(nreig[i],eigentex[i]);
      delete eigentex[i];
    }
    i++;
  }
  delete nreig;
  delete nruse;
  delete basisWidth;
  delete basisHeight;
  delete eigentex;
  delete colormatrix;
}

void PcaRendererNV::setDisplayList(int l) {
      glDeleteLists(pcalist,1);
      glNewList(pcalist,GL_COMPILE);
      (this->*setup_pca_state)();
      glCallList(l);
      glEndList();
}

void PcaRendererNV::setReduceFactor(int r)
{
  if(r<=0)
    r=1;
  reduceFactor=r;
  reinitTexture=1;
  /*
  largestBasisHeight=0;
  largestBasisWidth=0;
  while(channel<nrchannels){
    if(largestBasisHeight<basisHeight[channel])
      largestBasisHeight=basisHeight[channel];
    if(largestBasisWidth<basisWidth[channel])
      largestBasisWidth=basisWidth[channel];
    channel++;
    }*/
}

int PcaRendererNV::load(char *basepath,char **filenames)
{
  char s[512];
  
  for(int channel=0;channel<nrchannels;channel++)
  {
    strncpy(s,basepath,512);
    strncat(s,filenames[channel],256);
    if(!loadbasis(channel,s))
    {
      destroy();
      return 0;
    }
  }
}

int PcaRendererNV::initBasis(int channel,int width,int height,int numbasis)
{
  printDebug("initbasis NV\n");
  if(eigentex[channel]){
    glDeleteTextures(nreig[channel],eigentex[channel]);
    delete eigentex[channel];
  }
  printDebug("done freeing eigentex\n");
  basisWidth[channel]=width;
  basisHeight[channel]=height;
  nreig[channel]=numbasis;
  nruse[channel]=0;
  eigentex[channel]=new GLuint[numbasis];
  glGenTextures(nreig[channel],eigentex[channel]);

  if(channel==0)
    largestBasisWidth=largestBasisHeight=0;  
  if(basisWidth[channel]>largestBasisWidth)
    largestBasisWidth=basisWidth[channel];
  if(basisHeight[channel]>largestBasisHeight)
    largestBasisHeight=basisHeight[channel];

  return 1;
}

int PcaRendererNV::loadbasis(int channel,int start, int howmany,gzFile gzfile)
{
  char * data, * d;
  int i=start/4;
  int end=howmany+start;
  d=data=new char[basisWidth[channel]*basisWidth[channel]*howmany];

  printDebug("%d %d %d\n",basisWidth[channel],basisHeight[channel],howmany);
  
  nruse[channel]=start+howmany;
  
  gzread(gzfile,data,basisWidth[channel]*basisHeight[channel]*howmany);

  while(i<end/4){
    glBindTexture(GL_TEXTURE_2D,eigentex[channel][i]);

    inittex(d,basisWidth[channel],basisHeight[channel],GL_RGBA);
    d+=basisWidth[channel]*basisHeight[channel]*4;
    i++;
  }
  delete [] data;
  return 1;
}

int PcaRendererNV::loadbasis(int channel,char *filename)
{
  FILE *f;
  if(!(f=fopen(filename,"rb"))) return 0;
  else return loadbasis(channel,f);
}

int PcaRendererNV::loadbasis(int channel,gzFile  f)
{
  int texwidth,texheight,i;
  char *data,*d;
  
  gzread(f,&texwidth,4);
  gzread(f,&texheight,4);
  gzread(f,&nreig[channel],4);

  initBasis(channel,texwidth,texheight,nreig[channel]);

  return loadbasis(channel,0,nreig[channel],f);
}

void PcaRendererNV::setup_pca_state_NV20()
{
  glEnable(GL_REGISTER_COMBINERS_NV);
  glActiveTextureARB(GL_TEXTURE0_ARB);
  glEnable(GL_TEXTURE_2D);
  glActiveTextureARB(GL_TEXTURE1_ARB);
  glEnable(GL_TEXTURE_2D);
  glActiveTextureARB(GL_TEXTURE2_ARB);
  glEnable(GL_TEXTURE_2D);
  glActiveTextureARB(GL_TEXTURE3_ARB);
  glEnable(GL_TEXTURE_2D);
  
  glCombinerParameteriNV(GL_NUM_GENERAL_COMBINERS_NV,6);

  // tex0=tex0 . ccc0
  // tex1=tex1 . ccc1
  // spare0.a=tex0.a*ccc0.a + tex1.a*ccc1.a
  glCombinerInputNV(GL_COMBINER0_NV, GL_RGB, GL_VARIABLE_A_NV, GL_CONSTANT_COLOR0_NV,   GL_EXPAND_NORMAL_NV, GL_RGB);
  glCombinerInputNV(GL_COMBINER0_NV, GL_RGB, GL_VARIABLE_B_NV, GL_TEXTURE0,       GL_EXPAND_NORMAL_NV, GL_RGB);
  glCombinerInputNV(GL_COMBINER0_NV, GL_RGB, GL_VARIABLE_C_NV, GL_CONSTANT_COLOR1_NV, GL_EXPAND_NORMAL_NV, GL_RGB);
  glCombinerInputNV(GL_COMBINER0_NV, GL_RGB, GL_VARIABLE_D_NV, GL_TEXTURE1,       GL_EXPAND_NORMAL_NV, GL_RGB);
  
  glCombinerInputNV(GL_COMBINER0_NV, GL_ALPHA, GL_VARIABLE_A_NV, GL_CONSTANT_COLOR0_NV,   GL_EXPAND_NORMAL_NV, GL_ALPHA);
  glCombinerInputNV(GL_COMBINER0_NV, GL_ALPHA, GL_VARIABLE_B_NV, GL_TEXTURE0,       GL_EXPAND_NORMAL_NV, GL_ALPHA);
  glCombinerInputNV(GL_COMBINER0_NV, GL_ALPHA, GL_VARIABLE_C_NV, GL_CONSTANT_COLOR1_NV, GL_EXPAND_NORMAL_NV, GL_ALPHA);
  glCombinerInputNV(GL_COMBINER0_NV, GL_ALPHA, GL_VARIABLE_D_NV, GL_TEXTURE1,       GL_EXPAND_NORMAL_NV, GL_ALPHA);

  glCombinerOutputNV(GL_COMBINER0_NV, GL_RGB,GL_TEXTURE0, GL_TEXTURE1, GL_DISCARD_NV,GL_NONE,GL_NONE,GL_TRUE, GL_TRUE, GL_FALSE);
  glCombinerOutputNV(GL_COMBINER0_NV, GL_ALPHA,GL_DISCARD_NV, GL_DISCARD_NV, GL_TEXTURE0,GL_NONE,GL_NONE,GL_FALSE, GL_FALSE, GL_FALSE);

  printerror("rc0");

  // tex2=tex2 . ccc0
  // tex3=tex3 . ccc1
  // spare1.a=tex2.a*ccc0.a + tex3.a*ccc1.a
  glCombinerInputNV(GL_COMBINER1_NV, GL_RGB, GL_VARIABLE_A_NV, GL_CONSTANT_COLOR0_NV,   GL_EXPAND_NORMAL_NV, GL_RGB);
  glCombinerInputNV(GL_COMBINER1_NV, GL_RGB, GL_VARIABLE_B_NV, GL_TEXTURE2, GL_EXPAND_NORMAL_NV, GL_RGB);
  glCombinerInputNV(GL_COMBINER1_NV, GL_RGB, GL_VARIABLE_C_NV, GL_CONSTANT_COLOR1_NV, GL_EXPAND_NORMAL_NV, GL_RGB);
  glCombinerInputNV(GL_COMBINER1_NV, GL_RGB, GL_VARIABLE_D_NV, GL_TEXTURE3, GL_EXPAND_NORMAL_NV, GL_RGB);
  
  glCombinerInputNV(GL_COMBINER1_NV, GL_ALPHA, GL_VARIABLE_A_NV, GL_CONSTANT_COLOR0_NV,   GL_EXPAND_NORMAL_NV, GL_ALPHA);
  glCombinerInputNV(GL_COMBINER1_NV, GL_ALPHA, GL_VARIABLE_B_NV, GL_TEXTURE2, GL_EXPAND_NORMAL_NV, GL_ALPHA);
  glCombinerInputNV(GL_COMBINER1_NV, GL_ALPHA, GL_VARIABLE_C_NV, GL_CONSTANT_COLOR1_NV, GL_EXPAND_NORMAL_NV, GL_ALPHA);
  glCombinerInputNV(GL_COMBINER1_NV, GL_ALPHA, GL_VARIABLE_D_NV, GL_TEXTURE3, GL_EXPAND_NORMAL_NV, GL_ALPHA);
  
  glCombinerOutputNV(GL_COMBINER1_NV, GL_RGB,GL_TEXTURE2, GL_TEXTURE3, GL_DISCARD_NV,GL_NONE,GL_NONE,GL_TRUE, GL_TRUE, GL_FALSE);
  glCombinerOutputNV(GL_COMBINER1_NV,GL_ALPHA,GL_DISCARD_NV,GL_DISCARD_NV,GL_TEXTURE2,GL_NONE,GL_NONE,GL_FALSE,GL_FALSE, GL_FALSE);

  printerror("rc1");

  // spare0=tex1+tex2
  glCombinerInputNV(GL_COMBINER2_NV, GL_RGB, GL_VARIABLE_A_NV, GL_ZERO, GL_UNSIGNED_INVERT_NV, GL_RGB); 
  glCombinerInputNV(GL_COMBINER2_NV, GL_RGB, GL_VARIABLE_B_NV, GL_TEXTURE0,    GL_SIGNED_IDENTITY_NV,   GL_RGB);
  glCombinerInputNV(GL_COMBINER2_NV, GL_RGB, GL_VARIABLE_C_NV, GL_ZERO, GL_UNSIGNED_INVERT_NV, GL_RGB);
  glCombinerInputNV(GL_COMBINER2_NV, GL_RGB, GL_VARIABLE_D_NV, GL_TEXTURE1,    GL_SIGNED_IDENTITY_NV,   GL_RGB);
  
  glCombinerOutputNV(GL_COMBINER2_NV, GL_RGB, GL_DISCARD_NV, GL_DISCARD_NV, GL_SPARE0_NV,GL_NONE,GL_NONE,GL_FALSE,GL_FALSE, GL_FALSE);

  printerror("rc2");

  // spare1=tex3+tex4
  glCombinerInputNV(GL_COMBINER3_NV, GL_RGB, GL_VARIABLE_A_NV, GL_ZERO, GL_UNSIGNED_INVERT_NV, GL_RGB); 
  glCombinerInputNV(GL_COMBINER3_NV, GL_RGB, GL_VARIABLE_B_NV, GL_TEXTURE2,    GL_SIGNED_IDENTITY_NV,   GL_RGB);
  glCombinerInputNV(GL_COMBINER3_NV, GL_RGB, GL_VARIABLE_C_NV, GL_ZERO, GL_UNSIGNED_INVERT_NV, GL_RGB);
  glCombinerInputNV(GL_COMBINER3_NV, GL_RGB, GL_VARIABLE_D_NV, GL_TEXTURE3,    GL_SIGNED_IDENTITY_NV,   GL_RGB);
  
  glCombinerOutputNV(GL_COMBINER3_NV, GL_RGB, GL_DISCARD_NV, GL_DISCARD_NV, GL_SPARE1_NV,GL_NONE,GL_NONE,GL_FALSE,GL_FALSE, GL_FALSE);

  printerror("rc3");

  // spare0.rgb=spare0.rgb+spare1.rgb
  // spare0.a=tex0.a+tex2.a
  glCombinerInputNV(GL_COMBINER4_NV, GL_RGB, GL_VARIABLE_A_NV, GL_ZERO, GL_UNSIGNED_INVERT_NV, GL_RGB); 
  glCombinerInputNV(GL_COMBINER4_NV, GL_RGB, GL_VARIABLE_B_NV, GL_SPARE0_NV,    GL_SIGNED_IDENTITY_NV,   GL_RGB);
  glCombinerInputNV(GL_COMBINER4_NV, GL_RGB, GL_VARIABLE_C_NV, GL_ZERO, GL_UNSIGNED_INVERT_NV, GL_RGB);
  glCombinerInputNV(GL_COMBINER4_NV, GL_RGB, GL_VARIABLE_D_NV, GL_SPARE1_NV,    GL_SIGNED_IDENTITY_NV,   GL_RGB);

  glCombinerInputNV(GL_COMBINER4_NV, GL_ALPHA, GL_VARIABLE_A_NV, GL_ZERO, GL_UNSIGNED_INVERT_NV, GL_ALPHA); 
  glCombinerInputNV(GL_COMBINER4_NV, GL_ALPHA, GL_VARIABLE_B_NV, GL_TEXTURE0,    GL_SIGNED_IDENTITY_NV,   GL_ALPHA);
  glCombinerInputNV(GL_COMBINER4_NV, GL_ALPHA, GL_VARIABLE_C_NV, GL_ZERO, GL_UNSIGNED_INVERT_NV, GL_ALPHA);
  glCombinerInputNV(GL_COMBINER4_NV, GL_ALPHA, GL_VARIABLE_D_NV, GL_TEXTURE2,    GL_SIGNED_IDENTITY_NV,   GL_ALPHA);
  
  glCombinerOutputNV(GL_COMBINER4_NV, GL_RGB, GL_DISCARD_NV, GL_DISCARD_NV, GL_SPARE0_NV,GL_NONE,GL_NONE,GL_FALSE,GL_FALSE, GL_FALSE);
  glCombinerOutputNV(GL_COMBINER4_NV, GL_ALPHA, GL_DISCARD_NV, GL_DISCARD_NV, GL_SPARE0_NV,GL_NONE,GL_NONE,GL_FALSE,GL_FALSE, GL_FALSE);

  printerror("rc4");

  // spare0.rgb=spare0.rgb+spare0.a
  glCombinerInputNV(GL_COMBINER5_NV, GL_RGB, GL_VARIABLE_A_NV, GL_ZERO, GL_UNSIGNED_INVERT_NV, GL_RGB); 
  glCombinerInputNV(GL_COMBINER5_NV, GL_RGB, GL_VARIABLE_B_NV, GL_SPARE0_NV,    GL_SIGNED_IDENTITY_NV,   GL_RGB);
  glCombinerInputNV(GL_COMBINER5_NV, GL_RGB, GL_VARIABLE_C_NV, GL_ZERO, GL_UNSIGNED_INVERT_NV, GL_RGB);
  glCombinerInputNV(GL_COMBINER5_NV, GL_RGB, GL_VARIABLE_D_NV, GL_SPARE0_NV,    GL_SIGNED_IDENTITY_NV,   GL_ALPHA);
  
  glCombinerOutputNV(GL_COMBINER5_NV, GL_RGB, GL_DISCARD_NV, GL_DISCARD_NV, GL_SPARE0_NV,GL_NONE,GL_NONE,GL_FALSE,GL_FALSE, GL_FALSE);

  printerror("rc5");

//  glCombinerInputNV(GL_COMBINER6_NV, GL_RGB, GL_VARIABLE_A_NV, GL_FOG, GL_EXPAND_NORMAL_NV, GL_RGB); 
//  glCombinerInputNV(GL_COMBINER6_NV, GL_RGB, GL_VARIABLE_B_NV, GL_SPARE0_NV,    GL_SIGNED_IDENTITY_NV,   GL_RGB);
  
//  glCombinerOutputNV(GL_COMBINER6_NV, GL_RGB, GL_SPARE0_NV, GL_DISCARD_NV, GL_DISCARD_NV,GL_NONE,GL_NONE,GL_FALSE,GL_FALSE, GL_FALSE);
//  printerror("rc6");

  glFinalCombinerInputNV(GL_VARIABLE_A_NV,GL_FOG, GL_UNSIGNED_IDENTITY_NV, GL_RGB);
  glFinalCombinerInputNV(GL_VARIABLE_B_NV,GL_SPARE0_NV,GL_UNSIGNED_IDENTITY_NV, GL_RGB);
  glFinalCombinerInputNV(GL_VARIABLE_C_NV,GL_ZERO, GL_UNSIGNED_IDENTITY_NV, GL_RGB);
  glFinalCombinerInputNV(GL_VARIABLE_D_NV,GL_ZERO, GL_UNSIGNED_IDENTITY_NV, GL_RGB);
  glFinalCombinerInputNV(GL_VARIABLE_G_NV,GL_ZERO, GL_UNSIGNED_INVERT_NV, GL_ALPHA);

  printerror("rcF");
}

inline float byteCoeffToFloat(char c)
{
  return ((float)c)/254.0f + 0.5f;
}

void PcaRendererNV::renderpca_NV20(char ** coeff,GLuint tex)
{
  int tind,i,channel;
  float temp[4]={0,0,0,0.5};
  float zero[4]={0,0,0,0};
  float temp2[4]={1,0,0,1};
  int vp[4];

  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  glGetIntegerv(GL_VIEWPORT,vp);
  glViewport(0,0,largestBasisWidth>>(reduceFactor-1),largestBasisHeight>>(reduceFactor-1));
  glOrtho(0,1,0,1,-1000,1000);
  glDisable(GL_DEPTH_TEST);

  glEnable(GL_REGISTER_COMBINERS_NV);
  glEnable(GL_BLEND);

  for(channel=0;channel<nrchannels;channel++)
  {
    glFogfv(GL_FOG_COLOR,colormatrix+4*channel);
    tind=0;
    i=0;
    glBlendFunc(GL_ONE,GL_ZERO);

    if(channel==0)
    {
      glActiveTextureARB(GL_TEXTURE0);
      glBindTexture(GL_TEXTURE_2D,eigentex[channel][tind++]);
      glActiveTextureARB(GL_TEXTURE1);
      glBindTexture(GL_TEXTURE_2D,eigentex[channel][tind++]);
      glActiveTextureARB(GL_TEXTURE2);
      glBindTexture(GL_TEXTURE_2D,eigentex[channel][tind++]);
      glActiveTextureARB(GL_TEXTURE3);
      glBindTexture(GL_TEXTURE_2D,eigentex[channel][tind++]);

      temp[0]=byteCoeffToFloat(coeff[channel][i]); //old coeff[channel][i]/2+0.5;
      temp[1]=byteCoeffToFloat(coeff[channel][i+1]);
      temp[2]=byteCoeffToFloat(coeff[channel][i+2]);
      temp[3]=byteCoeffToFloat(coeff[channel][i+3]);
      glCombinerStageParameterfvNV(GL_COMBINER0_NV,GL_CONSTANT_COLOR0_NV,temp);

      temp[0]=byteCoeffToFloat(coeff[channel][i+4]);
      temp[1]=byteCoeffToFloat(coeff[channel][i+5]);
      temp[2]=byteCoeffToFloat(coeff[channel][i+6]);
      temp[3]=byteCoeffToFloat(coeff[channel][i+7]);
      glCombinerStageParameterfvNV(GL_COMBINER0_NV,GL_CONSTANT_COLOR1_NV,temp);

      temp[0]=byteCoeffToFloat(coeff[channel][i+8]);
      temp[1]=byteCoeffToFloat(coeff[channel][i+9]);
      temp[2]=byteCoeffToFloat(coeff[channel][i+10]);
      temp[3]=byteCoeffToFloat(coeff[channel][i+11]);
      glCombinerStageParameterfvNV(GL_COMBINER1_NV,GL_CONSTANT_COLOR0_NV,temp);

      temp[0]=byteCoeffToFloat(coeff[channel][i+12]);
      temp[1]=byteCoeffToFloat(coeff[channel][i+13]);
      temp[2]=byteCoeffToFloat(coeff[channel][i+14]);
      temp[3]=byteCoeffToFloat(coeff[channel][i+15]);
      glCombinerStageParameterfvNV(GL_COMBINER1_NV,GL_CONSTANT_COLOR1_NV,temp);

      glCallList(pcalist);

      i+=16;
    }

    printerror("renderpca_nv20_1");
    
    glBlendFunc(GL_ONE,GL_ONE);
  
    for(;i<nruse[channel];i+=16)
    {
      glActiveTextureARB(GL_TEXTURE0);
      glBindTexture(GL_TEXTURE_2D,eigentex[channel][tind++]);
      glActiveTextureARB(GL_TEXTURE1);
      glBindTexture(GL_TEXTURE_2D,eigentex[channel][tind++]);
      glActiveTextureARB(GL_TEXTURE2);
      glBindTexture(GL_TEXTURE_2D,eigentex[channel][tind++]);
      glActiveTextureARB(GL_TEXTURE3);
      glBindTexture(GL_TEXTURE_2D,eigentex[channel][tind++]);
      
      temp[0]=byteCoeffToFloat(coeff[channel][i]); //old coeff[channel][i]/2+0.5;
      temp[1]=byteCoeffToFloat(coeff[channel][i+1]);
      temp[2]=byteCoeffToFloat(coeff[channel][i+2]);
      temp[3]=byteCoeffToFloat(coeff[channel][i+3]);
      glCombinerStageParameterfvNV(GL_COMBINER0_NV,GL_CONSTANT_COLOR0_NV,temp);

      temp[0]=byteCoeffToFloat(coeff[channel][i+4]); 
      temp[1]=byteCoeffToFloat(coeff[channel][i+5]);
      temp[2]=byteCoeffToFloat(coeff[channel][i+6]);
      temp[3]=byteCoeffToFloat(coeff[channel][i+7]);
      glCombinerStageParameterfvNV(GL_COMBINER0_NV,GL_CONSTANT_COLOR1_NV,temp);

      temp[0]=byteCoeffToFloat(coeff[channel][i+8]); 
      temp[1]=byteCoeffToFloat(coeff[channel][i+9]);
      temp[2]=byteCoeffToFloat(coeff[channel][i+10]);
      temp[3]=byteCoeffToFloat(coeff[channel][i+11]);
      glCombinerStageParameterfvNV(GL_COMBINER1_NV,GL_CONSTANT_COLOR0_NV,temp);

      temp[0]=byteCoeffToFloat(coeff[channel][i+12]); 
      temp[1]=byteCoeffToFloat(coeff[channel][i+13]);
      temp[2]=byteCoeffToFloat(coeff[channel][i+14]);
      temp[3]=byteCoeffToFloat(coeff[channel][i+15]);
      glCombinerStageParameterfvNV(GL_COMBINER1_NV,GL_CONSTANT_COLOR1_NV,temp);
     
      glBlendEquation(GL_FUNC_ADD);
      glCallList(pcalist);

      temp[0]=-byteCoeffToFloat(coeff[channel][i]); 
      temp[1]=-byteCoeffToFloat(coeff[channel][i+1]);
      temp[2]=-byteCoeffToFloat(coeff[channel][i+2]);
      temp[3]=-byteCoeffToFloat(coeff[channel][i+3]);
      glCombinerStageParameterfvNV(GL_COMBINER0_NV,GL_CONSTANT_COLOR0_NV,temp);

      temp[0]=-byteCoeffToFloat(coeff[channel][i+4]); 
      temp[1]=-byteCoeffToFloat(coeff[channel][i+5]);
      temp[2]=-byteCoeffToFloat(coeff[channel][i+6]);
      temp[3]=-byteCoeffToFloat(coeff[channel][i+7]);
      glCombinerStageParameterfvNV(GL_COMBINER0_NV,GL_CONSTANT_COLOR1_NV,temp);

      temp[0]=-byteCoeffToFloat(coeff[channel][i+8]); 
      temp[1]=-byteCoeffToFloat(coeff[channel][i+9]);
      temp[2]=-byteCoeffToFloat(coeff[channel][i+10]);
      temp[3]=-byteCoeffToFloat(coeff[channel][i+11]);
      glCombinerStageParameterfvNV(GL_COMBINER1_NV,GL_CONSTANT_COLOR0_NV,temp);

      temp[0]=-byteCoeffToFloat(coeff[channel][i+12]); 
      temp[1]=-byteCoeffToFloat(coeff[channel][i+13]);
      temp[2]=-byteCoeffToFloat(coeff[channel][i+14]);
      temp[3]=-byteCoeffToFloat(coeff[channel][i+15]);
      glCombinerStageParameterfvNV(GL_COMBINER1_NV,GL_CONSTANT_COLOR1_NV,temp);

      glBlendEquation(GL_FUNC_REVERSE_SUBTRACT);
      glCallList(pcalist);
    }
  } 
  glPopMatrix();
  glActiveTextureARB(GL_TEXTURE0);
  glBindTexture(GL_TEXTURE_2D,tex);
  
  if(reinitTexture){
    char * emptyImage;
    emptyImage = (char *)malloc((largestBasisWidth>>(reduceFactor-1))*(largestBasisHeight>>(reduceFactor-1))*4);
    reinitTexture=0;
    glTexImage2D(GL_TEXTURE_2D,0,
		 GL_RGB,(largestBasisWidth>>(reduceFactor-1)),
		 (largestBasisHeight>>(reduceFactor-1)),
		 0,GL_RGBA,GL_UNSIGNED_BYTE,emptyImage);
    free(emptyImage);
  }
  glCopyTexSubImage2D(GL_TEXTURE_2D,0,0,0,0,0,
		      largestBasisWidth>>(reduceFactor-1),
		      largestBasisHeight>>(reduceFactor-1));
  
  glViewport(vp[0],vp[1],vp[2],vp[3]);
  glDisable(GL_REGISTER_COMBINERS_NV);
  printerror("renderpca_nv20");
}

void PcaRendererNV::setup_pca_state_NV17()
{
  glEnable(GL_REGISTER_COMBINERS_NV);
  glActiveTextureARB(GL_TEXTURE0_ARB);
  glEnable(GL_TEXTURE_2D);
  glActiveTextureARB(GL_TEXTURE1_ARB);
  glEnable(GL_TEXTURE_2D);
/*  glActiveTextureARB(GL_TEXTURE2_ARB);
  glEnable(GL_TEXTURE_2D);
  glActiveTextureARB(GL_TEXTURE3_ARB);
  glEnable(GL_TEXTURE_2D);*/
 
  glCombinerParameteriNV(GL_NUM_GENERAL_COMBINERS_NV,2);

  // spare0=tex0 . ccc0
  // spare1=tex1 . ccc1
  // spare0.a=tex0.a*ccc0.a + tex1.a*ccc1.a
  glCombinerInputNV(GL_COMBINER0_NV, GL_RGB, GL_VARIABLE_A_NV, GL_CONSTANT_COLOR0_NV,   GL_EXPAND_NORMAL_NV, GL_RGB);
  glCombinerInputNV(GL_COMBINER0_NV, GL_RGB, GL_VARIABLE_B_NV, GL_TEXTURE0,       GL_EXPAND_NORMAL_NV, GL_RGB);
  glCombinerInputNV(GL_COMBINER0_NV, GL_RGB, GL_VARIABLE_C_NV, GL_CONSTANT_COLOR1_NV, GL_EXPAND_NORMAL_NV, GL_RGB);
  glCombinerInputNV(GL_COMBINER0_NV, GL_RGB, GL_VARIABLE_D_NV, GL_TEXTURE1,       GL_EXPAND_NORMAL_NV, GL_RGB);
  
  glCombinerInputNV(GL_COMBINER0_NV, GL_ALPHA, GL_VARIABLE_A_NV, GL_CONSTANT_COLOR0_NV,   GL_EXPAND_NORMAL_NV, GL_ALPHA);
  glCombinerInputNV(GL_COMBINER0_NV, GL_ALPHA, GL_VARIABLE_B_NV, GL_TEXTURE0,       GL_EXPAND_NORMAL_NV, GL_ALPHA);
  glCombinerInputNV(GL_COMBINER0_NV, GL_ALPHA, GL_VARIABLE_C_NV, GL_CONSTANT_COLOR1_NV, GL_EXPAND_NORMAL_NV, GL_ALPHA);
  glCombinerInputNV(GL_COMBINER0_NV, GL_ALPHA, GL_VARIABLE_D_NV, GL_TEXTURE1,       GL_EXPAND_NORMAL_NV, GL_ALPHA);

  glCombinerOutputNV(GL_COMBINER0_NV, GL_RGB,GL_SPARE0_NV, GL_SPARE1_NV, GL_DISCARD_NV,GL_NONE,GL_NONE,GL_TRUE, GL_TRUE, GL_FALSE);
  glCombinerOutputNV(GL_COMBINER0_NV, GL_ALPHA,GL_DISCARD_NV, GL_DISCARD_NV, GL_TEXTURE0,GL_NONE,GL_NONE,GL_FALSE, GL_FALSE, GL_FALSE);

  printerror("rc0");

  // spare0.rgb=spare0 + spare1
  glCombinerInputNV(GL_COMBINER1_NV, GL_RGB, GL_VARIABLE_A_NV, GL_ZERO,   GL_UNSIGNED_INVERT_NV, GL_RGB);
  glCombinerInputNV(GL_COMBINER1_NV, GL_RGB, GL_VARIABLE_B_NV, GL_SPARE0_NV, GL_SIGNED_IDENTITY_NV, GL_RGB);
  glCombinerInputNV(GL_COMBINER1_NV, GL_RGB, GL_VARIABLE_C_NV, GL_ZERO, GL_UNSIGNED_INVERT_NV, GL_RGB);
  glCombinerInputNV(GL_COMBINER1_NV, GL_RGB, GL_VARIABLE_D_NV, GL_SPARE1_NV, GL_SIGNED_IDENTITY_NV, GL_RGB);
  
  glCombinerOutputNV(GL_COMBINER1_NV, GL_RGB,GL_DISCARD_NV, GL_DISCARD_NV, GL_SPARE0_NV,GL_NONE,GL_NONE,GL_FALSE, GL_FALSE, GL_FALSE);

  printerror("rc1");

  glFinalCombinerInputNV(GL_VARIABLE_A_NV,GL_FOG, GL_UNSIGNED_IDENTITY_NV, GL_RGB);
  glFinalCombinerInputNV(GL_VARIABLE_B_NV,GL_TEXTURE0,GL_UNSIGNED_IDENTITY_NV, GL_ALPHA);
  glFinalCombinerInputNV(GL_VARIABLE_C_NV,GL_ZERO, GL_UNSIGNED_IDENTITY_NV, GL_RGB);
  glFinalCombinerInputNV(GL_VARIABLE_D_NV,GL_E_TIMES_F_NV,GL_UNSIGNED_IDENTITY_NV, GL_RGB);
  glFinalCombinerInputNV(GL_VARIABLE_E_NV,GL_SPARE0_NV,GL_UNSIGNED_IDENTITY_NV, GL_RGB);
  glFinalCombinerInputNV(GL_VARIABLE_F_NV,GL_FOG,GL_UNSIGNED_IDENTITY_NV, GL_RGB);
  glFinalCombinerInputNV(GL_VARIABLE_G_NV,GL_ZERO, GL_UNSIGNED_INVERT_NV, GL_ALPHA);

  printerror("rcF");
}

void PcaRendererNV::renderpca_NV17(char **coeff,GLuint tex)
{
  int tind,i,channel;
  int vp[4];
  float temp[4]={0,0,0,0.5};
  float zero[4]={0,0,0,0};

  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  glGetIntegerv(GL_VIEWPORT,vp);

  glViewport(0,0,largestBasisWidth>>(reduceFactor-1)
	     ,largestBasisHeight>>(reduceFactor-1));

  glOrtho(0,1,0,1,-1000,1000);
  glDisable(GL_DEPTH_TEST);

  glEnable(GL_BLEND);
  glEnable(GL_REGISTER_COMBINERS_NV);

  for(channel=0;channel<nrchannels;channel++)
  {
    glFogfv(GL_FOG_COLOR,colormatrix+4*channel);
    tind=0;

    i=0;
    
    glBlendFunc(GL_ONE,GL_ZERO);

    if(channel==0)
    {
      glActiveTextureARB(GL_TEXTURE0);
      glBindTexture(GL_TEXTURE_2D,eigentex[channel][tind++]);
      glActiveTextureARB(GL_TEXTURE1);
      glBindTexture(GL_TEXTURE_2D,eigentex[channel][tind++]);
      
      temp[0]=byteCoeffToFloat(coeff[channel][i]); 
      temp[1]=byteCoeffToFloat(coeff[channel][i+1]);
      temp[2]=byteCoeffToFloat(coeff[channel][i+2]);
      temp[3]=byteCoeffToFloat(coeff[channel][i+3]);
      glCombinerParameterfvNV(GL_CONSTANT_COLOR0_NV,temp);
      
      temp[0]=byteCoeffToFloat(coeff[channel][i+4]); 
      temp[1]=byteCoeffToFloat(coeff[channel][i+5]);
      temp[2]=byteCoeffToFloat(coeff[channel][i+6]);
      temp[3]=byteCoeffToFloat(coeff[channel][i+7]);
      glCombinerParameterfvNV(GL_CONSTANT_COLOR1_NV,temp);
     
      glCallList(pcalist);

      i+=8;
    }
    
    glBlendFunc(GL_ONE,GL_ONE);
  
    for(;i<nruse[channel];i+=8)
    {
      glActiveTextureARB(GL_TEXTURE0);
      glBindTexture(GL_TEXTURE_2D,eigentex[channel][tind++]);
      glActiveTextureARB(GL_TEXTURE1);
      glBindTexture(GL_TEXTURE_2D,eigentex[channel][tind++]);

      temp[0]=byteCoeffToFloat(coeff[channel][i]); 
      temp[1]=byteCoeffToFloat(coeff[channel][i+1]);
      temp[2]=byteCoeffToFloat(coeff[channel][i+2]);
      temp[3]=byteCoeffToFloat(coeff[channel][i+3]);
      glCombinerParameterfvNV(GL_CONSTANT_COLOR0_NV,temp);

      temp[0]=byteCoeffToFloat(coeff[channel][i+4]); 
      temp[1]=byteCoeffToFloat(coeff[channel][i+5]);
      temp[2]=byteCoeffToFloat(coeff[channel][i+6]);
      temp[3]=byteCoeffToFloat(coeff[channel][i+7]);
      glCombinerParameterfvNV(GL_CONSTANT_COLOR1_NV,temp);
     
      glBlendEquation(GL_FUNC_ADD);
      glCallList(pcalist);

      temp[0]=-byteCoeffToFloat(coeff[channel][i]); 
      temp[1]=-byteCoeffToFloat(coeff[channel][i+1]);
      temp[2]=-byteCoeffToFloat(coeff[channel][i+2]);
      temp[3]=-byteCoeffToFloat(coeff[channel][i+3]);
      glCombinerParameterfvNV(GL_CONSTANT_COLOR0_NV,temp);
      
      temp[0]=-byteCoeffToFloat(coeff[channel][i+4]); 
      temp[1]=-byteCoeffToFloat(coeff[channel][i+5]);
      temp[2]=-byteCoeffToFloat(coeff[channel][i+6]);
      temp[3]=-byteCoeffToFloat(coeff[channel][i+7]);
      glCombinerParameterfvNV(GL_CONSTANT_COLOR1_NV,temp);

      glBlendEquation(GL_FUNC_REVERSE_SUBTRACT);
      glCallList(pcalist);
    }
  } 
  glPopMatrix();
  glActiveTextureARB(GL_TEXTURE0);
  glBindTexture(GL_TEXTURE_2D,tex);
  if(reinitTexture){
    char * emptyImage;
    emptyImage = (char *)malloc((largestBasisWidth>>(reduceFactor-1))*(largestBasisHeight>>(reduceFactor-1))*4);
    reinitTexture=0;
    glTexImage2D(GL_TEXTURE_2D,0,
		 GL_RGB,(largestBasisWidth>>(reduceFactor-1)),(largestBasisHeight>>(reduceFactor-1)),
		 0,GL_RGBA,GL_UNSIGNED_BYTE,emptyImage);
    free(emptyImage);
  }
  glCopyTexSubImage2D(GL_TEXTURE_2D,0,0,0,0,0,largestBasisWidth>>(reduceFactor-1),largestBasisHeight>>(reduceFactor-1));
  glViewport(vp[0],vp[1],vp[2],vp[3]);

  glDisable(GL_REGISTER_COMBINERS_NV);
}

int PcaRendererSoft::numCreated=0;
int PcaRendererSoft::hasMMX=0;

int PcaRendererSoft::getMMX()
{
  int caps;
  __asm__("mov $1,%%eax\n\t"
	  "cpuid\n\t"
	  "mov %%edx,%0\n\t"
	  : "=m" (caps)
	  :
	  : "%eax","%edx","%ecx","ebx");
  PcaRendererSoft::hasMMX=(caps&0x800000)>0;
  return PcaRendererSoft::hasMMX;
}

PcaRendererSoft::PcaRendererSoft(int num_channels)
{
  if(PcaRendererSoft::numCreated==0){
    PcaRendererSoft::getMMX();
    PcaRendererSoft::numCreated++;
  }
  setUsingMMX(PcaRendererSoft::hasMMX);
  
  quality=1.0f;
  reduceFactor=1;
  
  initNumChannels(num_channels);

  basis = new (char *)[num_channels];
  basisWidth = new int[num_channels];
  basisHeight = new int[num_channels];

  memset(basis,0,sizeof(char *)*num_channels);
  memset(basisWidth,0,sizeof(int)*num_channels);
  memset(basisHeight,0,sizeof(int)*num_channels);

  currentBasis = new (char *)[num_channels];
  memset(currentBasis,0,sizeof(char *)*num_channels);

  currentBasisWidth = new int[num_channels];
  currentBasisHeight = new int[num_channels];
  memset(currentBasisWidth,0,sizeof(int)*num_channels);
  memset(currentBasisHeight,0,sizeof(int)*num_channels);
  

  largestBasisWidth  = 0;
  largestBasisHeight = 0;

  imageShort = new (short *)[num_channels];
  memset(imageShort,0,sizeof(short *)*num_channels);
  image=NULL;
  //glActiveTextureARB(GL_TEXTURE0_ARB);
  //glEnable(GL_TEXTURE_2D);
}

void PcaRendererSoft::destroy()
{
  //destroy
  if(currentBasis){
    int i=0;
    while(i<nrchannels){
      if(currentBasis[i]!=basis[i])
	delete currentBasis[i];
      i++;
    }
    delete currentBasis;
  }
  if(basis){
    int i=0;
    while(i<nrchannels){
      delete basis[i];
      basis[i]=NULL;
      i++;
    }
    delete basis;
    basis=NULL;
  }
  if(imageShort){
    int i=0;
    while(i<nrchannels){
      delete imageShort[i];
      i++;
    }
    delete imageShort;
  }
  if(image){
    delete image;
  }
  printDebug("destroying software\n");
  delete currentBasisHeight;
  delete currentBasisWidth;
  delete basisWidth;
  delete basisHeight;
  delete eigentex;
  delete colormatrix;
}

void PcaRendererSoft::setReduceFactor(int r)
{
  if(reduceFactor>1){
    int channel=0;
    while(channel<nrchannels){
      delete currentBasis[channel];
      channel++;
    }
  }
  if(r<=0)
    r=1;
  reduceFactor=r;
  if(reduceFactor>1){
    int channel=0;
    while(channel<nrchannels){
      int newwidth=currentBasisWidth[channel] = (basisWidth[channel]>>(r-1));
      int newheight=currentBasisHeight[channel] = (basisHeight[channel]>>(r-1));
      int size = newwidth*newheight*nreig[channel];
      int j=0;
      int jnew=0;
      int inew=0;
      int i=0;
      int increment = (1<<(reduceFactor-1));
      int k=0;
      currentBasis[channel] = new char[size];
     
      while(k<nreig[channel]){
	i=increment/2;
	inew=0;
	while(i<basisHeight[channel]){
	  j=increment/2;
	  jnew=0;
	  while(j<basisWidth[channel]){
	    currentBasis[channel][k*newwidth*newheight+newwidth*inew+jnew]=
	      basis[channel][k*basisWidth[channel]*basisHeight[channel]+basisWidth[channel]*i+j];
	    j+=increment;
	    jnew++;
	  }
	  i+=increment;
	  inew++;
	}
	k++;
      }
      channel++;
    }
  }
  /*
    The reduce factor is 1, so just use the original basis
  */
  else {
    int channel=0;
    while(channel<nrchannels){
      currentBasisWidth[channel]=basisWidth[channel];
      currentBasisHeight[channel]=basisHeight[channel];
      currentBasis[channel]=basis[channel];
      channel++;
    }
  }
  /* Update the largestBasisWidth and height */
  {
    int channel=0;
    largestBasisWidth=largestBasisHeight=0;
    while(channel<nrchannels){
      if(currentBasisWidth[channel] > largestBasisWidth){
	largestBasisWidth = currentBasisWidth[channel];
      }
      if(currentBasisHeight[channel] > largestBasisHeight){
	largestBasisHeight = currentBasisHeight[channel];
      }
      channel++;
    }
  }
}

int PcaRendererSoft::load(char * basepath,char **filenames)
{
  char s[512];
  
  for(int channel=0;channel<nrchannels;channel++)
    {
      strncpy(s,basepath,512);
      strncat(s,filenames[channel],256);
      if(!loadbasis(channel,s))
	{
	  destroy();
	  return 0;
	}
    }
  
  printDebug("in pcarenderer load\n");
  return 1;
}

int PcaRendererSoft::initBasis(int channel,int width,int height,int numbasis)
{  
  int i=0;

  if(imageShort[channel])
    delete imageShort[channel];
  if(basis[channel])
    delete basis[channel];

  imageShort[channel] = new short[width*height];
  basis[channel]= new char[width*height*numbasis];
  
  basisWidth[channel]=width;
  basisHeight[channel]=height;
  nreig[channel]=numbasis;
  nruse[channel]=0;

  currentBasis[channel]=basis[channel];
  currentBasisWidth[channel]=basisWidth[channel];
  currentBasisHeight[channel]=basisHeight[channel];

  largestBasisWidth=0;
  largestBasisHeight=0;
  /*
    Since we are not garunteed that these calls will be made in 
    any order, so we cannot reset the largestBasisWidth at any
    time.
  */
  while(i<nrchannels){
    if(basisWidth[i]>largestBasisWidth)
      largestBasisWidth=basisWidth[i];
    if(basisHeight[i]>largestBasisHeight)
      largestBasisHeight=basisHeight[i];
    i++;
  }	  
  
  
  //FIXME: this must be called after all basis are loaded
  if(channel==nrchannels-1){
    if(image)
      delete [] image;
    image = new rgb[largestBasisWidth*largestBasisHeight];
  }
  return 1;
}

int PcaRendererSoft::loadbasis(int channel,int start,int howmany,gzFile f)
{
  int i=start;
  int end=start+howmany;
  char * d,* data;

  nruse[channel]=start+howmany;

  int texwidth=basisWidth[channel];
  int texheight=basisHeight[channel];
  d=data=new char[texwidth*texheight*howmany];
  if(gzread(f,data,texwidth*texheight*howmany)<=0)
    return 0;


  for(i=start;i<end;i+=4)
  {
    printDebug("calling inittex for channel: %d,nreig[channel]: %d\n",
	       channel,
	       nreig[channel]);

    int j=0;
    int vectorSize=texwidth*texheight;
    for(j=0;j<vectorSize;j++){
      basis[channel][i*(vectorSize)+j]=    d[0]-128;
      basis[channel][(i+1)*(vectorSize)+j]=d[1]-128;
      basis[channel][(i+2)*(vectorSize)+j]=d[2]-128;
      basis[channel][(i+3)*(vectorSize)+j]=d[3]-128;
      d+=4;
    }
  }  
  delete data;
  return 1;
}

int PcaRendererSoft::loadbasis(int channel, char * filename)
{
  gzFile f;
  if(!(f=gzopen(filename,"rb"))){
    return 0;
  }
  else{
    int ret= loadbasis(channel,f);
    gzclose(f);
    return ret;
  }
}

int PcaRendererSoft::loadbasis(int channel,gzFile  f)
{
  int texwidth,texheight;

  gzread(f,&texwidth,4);
  gzread(f,&texheight,4);
  gzread(f,&nreig[channel],4);
  
  initBasis(channel,texwidth,texheight,nreig[channel]);
  return loadbasis(channel,0,nreig[channel],f);
}

//FIXME: These should be dynamic not static
int ci[3][128];

void PcaRendererSoft::renderpcaMMX2()
{
  int i=0;
  int j=0;
  int channel=0;
  int top;
  __asm__("emms");
  while(channel<nrchannels){
    top = currentBasisHeight[channel]*currentBasisWidth[channel];
    i=0;
   
    while(i<nruse[channel]){
      int c1=ci[channel][i];
      int c2=ci[channel][i+1];
      short tc[2];
      char * basis1 = &(currentBasis[channel][i
				      *currentBasisWidth[channel]
				      *currentBasisHeight[channel]]);
      char * basis2 = basis1+top;

      tc[0]=c1;
      tc[1]=c2;
      tc[2]=c1;
      tc[3]=c2;
      /* set up the coefficients */
      __asm__("movq %0,%%mm7"
	      : 
	      : "m" (tc[0]));
      __asm__("punpckldq %mm7,%mm7"); 
      
      j=0;
      while(j<top){
	__asm__("movd %0,%%mm0"
		:
		:"m" (basis1[j]));
	__asm__("movd %0,%%mm1"
		:
		:"m" (basis2[j]));
	//Now the basis is in mm0 and mm1
	  
	//unpackand sign extend the bytes into words
	//So basis1 is now in mm2
	__asm__("pxor %mm2,%mm2");
	__asm__("punpcklbw %mm0,%mm2");
	__asm__("psraw $8,%mm2");

	//Do the same for mm1, so the basis2 is now
	//in mm3
	__asm__("pxor %mm3,%mm3");
	__asm__("punpcklbw %mm1,%mm3");
	__asm__("psraw $8,%mm3");

	//Read in the image into mm1, since mm1 is free now
	__asm__("movq %0,%%mm1"
		:
		: "m" (imageShort[channel][j]));
  
	//before interleaving the basis'
	//Copy mm2 over to mm4
	__asm__("movq %mm2,%mm4");
	__asm__("punpcklwd %mm3,%mm2");
	__asm__("punpckhwd %mm3,%mm4");
	//Now mm2 containes the b2p2 | b1p2 | b2p1 | b1p1
	//Now mm4 contains b2p4 | b1p4 | b2p3 | b1p3
	
	//multiply and add the coefficients
	__asm__("pmaddwd  %mm7,%mm2");
	// mm2 contains b2p2*c2+p1p2*c1 | b2p1*c2+b2p1+c2
	
	//multiply and add the higher pixels
	__asm__("pmaddwd  %mm7,%mm4");
	// mm4 contains b2p4*c2+p1p4*c1 | b2p3*c2+b2p3+c2
   
	//Clear register 0
	__asm__("pxor %mm0,%mm0");
	//Pack the first two pixels into the low end of mm2 as words
	__asm__("packssdw %mm0,%mm2");
	//Pack the other two pixels into thw high end of mm0 as words
	__asm__("packssdw %mm4,%mm0");
	
	/* add first two pixels back */
	__asm__("paddsw %mm2,%mm1");
	/* add last two pixels back */
	__asm__("paddsw %mm0,%mm1");
	//Store back ine image
	__asm__("movq   %%mm1,%0"
		: "=m" (imageShort[channel][j]));
	j+=4;
      }
      i+=2;
    }
    channel++;
  }
  __asm__("emms");
}

void PcaRendererSoft::renderpcaNoMMX()
{
  int i=0;
  int j=0;
  int top;
  int channel=0;   
  while(channel<nrchannels){
    top = currentBasisHeight[channel]*currentBasisWidth[channel];
    i=0;
    while(i<nruse[channel]){
      char c=ci[channel][i];
      char * basisImage = &(currentBasis[channel][i*top]);
      j=0;
      while(j<top){
	imageShort[channel][j]=imageShort[channel][j]+(c*basisImage[j]);
	j++;
      }
      i++;
    }
    channel++;
  }
}

void PcaRendererSoft:: renderpcaMMX()
{
  int i=0;
  int j=0;
  int channel=0;
  int top;
  
  __asm__("emms");
  __asm__("pxor   %mm0, %mm0");
  
  while(channel<nrchannels){
    top = currentBasisHeight[channel]*currentBasisWidth[channel];
    i=0;
   
    while(i<nruse[channel]){
      int c=ci[channel][i];
      //      c = c&0xFFFF;
      char * basisImage = &(currentBasis[channel][i*top]);
      __asm__("movd  %0,%%mm1"
	      : 
	      : "m" (ci[channel][i]));
      __asm__("punpcklwd  %mm1,%mm1");
      __asm__("punpckldq  %mm1,%mm1");
      
      __asm__("nop");
      j=0;
      while(j<top){
	__asm__("movq %0,%%mm7"
		:
		: "m" (imageShort[channel][j]));
	__asm__("movd %0,%%mm0"
		:
		: "m" (basisImage[j]));
	
	__asm__("pxor  %mm5,%mm5");
	__asm__("punpcklbw  %mm0,%mm5");
	__asm__("psraw      $8,%mm5");
	__asm__("pmullw     %mm1,%mm5");
	//Is this multiplication doing what we want
	
	/* add the quad word from memory to the
	   new result, then write the new sum back to
	   memory
	*/
	__asm__("paddw %mm5,%mm7");
	
	/*original
	  __asm__("movq %0,%%mm7"
	  :
	  : "m" (imageShort[channel][j]));
	  __asm__("movd %0,%%mm5"
		  :
		: "m" (basisImage[j]));
	__asm__("punpcklbw  %mm0,%mm5");
	__asm__("pmullw     %mm1,%mm5");
	__asm__("paddusw %mm5,%mm7");
  
	  __asm__("movd %0,%%mm5"
		  :
		  : "m" (basis2[k]));
	  __asm__("punpcklbw  %mm0,%mm5");
	  __asm__("pmullw     %mm2,%mm5");
	  __asm__("paddw %mm5,%mm7");
	  
	  __asm__("movd %0,%%mm5"
		  :
		  : "m" (basis3[k]));
	  __asm__("punpcklbw  %mm0,%mm5");
	  __asm__("pmullw     %mm3,%mm5");
	  __asm__("paddw %mm5,%mm7");
	  
	  __asm__("movd %0,%%mm5"
		  :
		  : "m" (basis4[k]));
	  __asm__("punpcklbw  %mm0,%mm5");
	  __asm__("pmullw     %mm4,%mm5");
	  __asm__("paddw %mm5,%mm7");
	  */
	  
	__asm__("movq  %%mm7,%0"
		: "=m" (imageShort[channel][j]));
	
	j+=4;
	
	//imageShort[channel][j]=imageShort[channel][j]+(c*basisImage[j]);
	//j++;
      }
      i++;
    }
    channel++;
  }
  __asm__("emms");
}

void printReg(int reg)
{
  char r[8];
  char cmd[16];
  int i=0;
  memset(r,0,8);
  printf("mm%d:",reg);
  switch(reg){
  case 0:
    __asm__("movq %%mm0,%0"
	    : "=m" (r[0]));break;
  case 1:
    __asm__("movq %%mm1,%0"
	    : "=m" (r[0]));break;
  case 2:
    __asm__("movq %%mm2,%0"
	  
	    : "=m" (r[0]));break;
  case 3:
    __asm__("movq %%mm3,%0"
	  
	    : "=m" (r[0]));break;
  case 4:
    __asm__("movq %%mm4,%0"
	  
	    : "=m" (r[0]));break;
  case 5:
    __asm__("movq %%mm5,%0"
	  
	    : "=m" (r[0]));break;
  case 6:
    __asm__("movq %%mm6,%0"
	    : "=m" (r[0]));break;
  case 7:
    __asm__("movq %%mm7,%0"
	    : "=m" (r[0]));break;
  }
  i=7;
  while(i>=0){
    printf("[%3d]",r[i]);
    i--;
  }
  putchar(10);
}

/*
  void PcaRenderSoft::applyColorMatrixNoMMX()
  pre: the images data has been filled and just needs
       to be multiplied by the color matrix
  post: The resulting image (the unsigned char) will be
        set to the original image multiplied by the 
	current colorMatrix
 
  limitations:  Assumes that the pre-colormatrix textures 
  are always smaller than the resulting rendered image.
  This is so they can be stretched efficiently in this loop
  as well.
  Currently the x increment assumes that the these texture sizes
  are all smaller than 256.  To add support for wider 
  textures, simply change the x increment to increment
  each channel at once.
*/   
void PcaRendererSoft::applyColorMatrixNoMMX()
{
 
  short * imagesy, *imagesu,* imagesv;
  int i=0;
  int j=0;

  short tempColor[4];
  unsigned char xsft[4];
  unsigned char ysft[4];
  short f;

  i=0;
  while(i<nrchannels){
    xsft[i]=(largestBasisWidth/currentBasisWidth[i])/2;
    ysft[i]=(largestBasisHeight/currentBasisHeight[i])/2;
    i++;
  }

  i=0;
  while(i<largestBasisHeight){
    j=0;
    
    imagesy=imageShort[0]+(i>>ysft[0])*currentBasisWidth[0];
    imagesu=imageShort[1]+(i>>ysft[1])*currentBasisWidth[1];
    imagesv=imageShort[2]+(i>>ysft[2])*currentBasisWidth[2];
    
    while(j<largestBasisWidth){
      int imindex=(i*largestBasisWidth)+j;
      //printf("(%d,%d)= \n",i,j,imindex);
      tempColor[0]=imagesy[j>>xsft[0]]>>6;
      tempColor[1]=imagesu[j>>xsft[1]]>>6;
      tempColor[2]=imagesv[j>>xsft[2]]>>6;
      tempColor[3]=255;//FIXME use the other channel!!

      f=  
	colorMatrixShort[0]*tempColor[0] 
	+ colorMatrixShort[1]*tempColor[1] 
	+ colorMatrixShort[2]*tempColor[2];
      
      f=f>>6;
      
      if(f<0)
	image[imindex].r=0;
      else if(f>255)
	image[imindex].r=255;
      else 
	image[imindex].r=f;
      
      f= colorMatrixShort[4]*tempColor[0] 
	+colorMatrixShort[5]*tempColor[1]
	+colorMatrixShort[6]*tempColor[2];
      
      f=f>>6;
      if(f<0)
	image[imindex].g=0;
      else if(f>255)
	image[imindex].g=255;
      else
	image[imindex].g=f;
      
      f= colorMatrixShort[8]*tempColor[0]
	+colorMatrixShort[9]*tempColor[1]+
	+colorMatrixShort[10]*tempColor[2];
      
      f=f>>6;      
      if(f<0)
	image[imindex].b=0;
      else if(f>255)
	image[imindex].b=255;
      else
	image[imindex].b =f;

      image[imindex].a=255;

      //OLD MATRIX
      //r=2y 0 1.402v
      //g=2y -0.3444136 -0.7416v
      //b=2y +1.773 0
      j++;
    }
    i++;
  }
}

/*
  void PcaRenderSoft::applyColorMatrixNoMMX()
  pre: the images data has been filled and just needs
       to be multiplied by the color matrix
  post: The resulting image (the unsigned char) will be
        set to the original image multiplied by the 
	current colorMatrix.  The multiplications are
	carried out using MMX
 
  limitations:  
  Assumes that the pre-colormatrix textures 
  are always smaller than the resulting rendered image.
  This is so they can be stretched efficiently in this loop
  as well.
  
  So far this function has not proven to increase 
  efficiency that much more than the normal implementation
  but may on some computers, and is never any slower.
*/
void PcaRendererSoft::applyColorMatrixMMX()
{
  short * imagesy, *imagesu,* imagesv;
  int i=0;
  int j=0;
  short mm0[4];
  short mm1[4];
  short mm2[4];
  short mm3[4];

  short tempColor[4];
  unsigned char xrsft[4];
  unsigned char yrsft[4];
  unsigned char xlsft[4];
  unsigned char ylsft[4];
  /*
    Since the textures are smaller, figure out
    how much we must shift the current counter
    in order to get the index into the smaller image.
  */
  i=0;
  while(i<nrchannels){
    xrsft[i]=(largestBasisWidth/currentBasisWidth[i])/2;
    yrsft[i]=(largestBasisHeight/currentBasisHeight[i])/2;
    i++;
  }

  /*
    Set up the coefficients for the color matrix
    in appropriate MMX registers.
    Do this in variables since they are easier
    to work with
  */
  mm0[0]=colorMatrixShort[0];
  mm0[1]=colorMatrixShort[1];
  mm0[2]=colorMatrixShort[4];
  mm0[3]=colorMatrixShort[5];
  mm1[0]=colorMatrixShort[2];
  mm1[1]=colorMatrixShort[3];
  mm1[2]=colorMatrixShort[6];
  mm1[3]=colorMatrixShort[7];
  mm2[0]=colorMatrixShort[8];
  mm2[1]=colorMatrixShort[9];
  mm2[2]=colorMatrixShort[12];
  mm2[3]=colorMatrixShort[13];
  mm3[0]=colorMatrixShort[10];
  mm3[1]=colorMatrixShort[11];
  mm3[2]=colorMatrixShort[14];
  mm3[3]=colorMatrixShort[15];

  __asm__("emms");
  __asm__("movq %0,%%mm0\n\t"
	  "movq %1,%%mm1\n\t"
	  "movq %2,%%mm2\n\t"
	  "movq %3,%%mm3\n\t"
	  :
	  : "m" (mm0[0]),
	  "m" (mm1[0]),
 	  "m" (mm2[0]),
	  "m" (mm3[0]));
  i=0;
  while(i<largestBasisHeight){
    j=0;
    
    
    imagesy=imageShort[0]+
      ((i>>yrsft[0]))*currentBasisWidth[0];
    imagesu=imageShort[1]+
      ((i>>yrsft[1]))*currentBasisWidth[1];
    imagesv=imageShort[2]+
      ((i>>yrsft[2]))*currentBasisWidth[2];

    while(j<largestBasisHeight){
      int imindex=(i*largestBasisWidth)+j;

      tempColor[0]=
	imagesy[(j>>xrsft[0])]>>6;
      tempColor[1]=
	imagesu[(j>>xrsft[1])]>>6;
      tempColor[2]=
	imagesv[(j>>xrsft[2])]>>6;
      tempColor[3]=255;//FIXME: use the other channel!!

      short tempshort[4];
      unsigned char tempchar[4];
      __asm__("movd %0,%%mm4\n\t"
	      "punpckldq %%mm4,%%mm4\n\t"
	      "movq  %%mm4,%%mm6\n\t"
	      "movd %1,%%mm5\n\t"
	      "punpckldq %%mm5,%%mm5\n\t"
	      "movq  %%mm5,%%mm7\n\t"
	      :
	      : "m" (tempColor[0]),
	      "m" (tempColor[2]));
     
      __asm__("pmaddwd  %mm0,%mm4");
      __asm__("pmaddwd  %mm1,%mm5");
      __asm__("paddd %mm4,%mm5");

      __asm__("pmaddwd  %mm2,%mm6");
      __asm__("pmaddwd  %mm3,%mm7");
      __asm__("paddd %mm6,%mm7");
      __asm__("psrad  $6,%mm5");

      __asm__("psrad  $6,%mm7");
      __asm__("pxor %mm4,%mm4");
            
      __asm__("packssdw %mm4,%mm5");

      
      __asm__("packssdw %mm7,%mm4");
      __asm__("por   %mm5,%mm4");
      __asm__("pxor  %mm6,%mm6");

      __asm__("packuswb %mm6,%mm4");      

      __asm__("movd %%mm4,%0"
	      : "=m"(image[imindex]));
      j++;
    }
    i++;
  }
  __asm__("emms");
}

/*
  void PcaRenderSoft::renderpca(float ** coeff)
  pre: the coefficients are all set, and are approximately
       between -1 and 1
  post: Use the coefficients and the current basis to
        create an image by multiplying each basis by the
	appropriate coefficient
*/
void PcaRendererSoft::renderpca(float ** coeff,GLuint tex)
{
  /*
    All the processing occurs in bytes, so scale the 
    coefficients so that they all fit into a byte,
    and correct after the image has been rendered
  */
  int channel=0;
  int i=0;
  while(channel<nrchannels){
    i=0;
    while(i<nreig[channel]){
    
      ci[channel][i]=((char)(coeff[channel][i]*127.0f));
      i++;
    }
    channel++;
  }
  renderpcabody(tex);
}

/*
  void PcaRenderSoft::renderpca(char ** coeff,GLuint tex)
  pre: the coefficients are all set, and are approximately
       between -1 and 1
  post: Use the coefficients and the current basis to
        create an image by multiplying each basis by the
	appropriate coefficient
	
*/
void PcaRendererSoft::renderpca(char ** coeff,GLuint tex)
{
  /*
    All the processing occurs in bytes, so scale the 
    coefficients so that they all fit into a byte,
    and correct after the image has been rendered
  */
  int channel=0;
  int i=0;
  while(channel<nrchannels){
    i=0;
    while(i<nreig[channel]){
      
      ci[channel][i]=(coeff[channel][i]);
      i++;
    }
    channel++;
  }
  renderpcabody(tex);
}

void PcaRendererSoft::renderpcabody(GLuint tex)
{
  int i=0;
  int j=0;
  int top=256*256;
    
  memset(colorMatrixShort,0,sizeof(char)*16);
  colorMatrixShort[0]=
    colorMatrixShort[4]=
    colorMatrixShort[8]=64;
  colorMatrixShort[2]=64;
  colorMatrixShort[9]=64;
  colorMatrixShort[15]=64;

  /* Clear the old image short buffer */
  i=0;
  while(i<nrchannels){
    memset(imageShort[i],0,sizeof(short)*currentBasisHeight[i]*currentBasisWidth[i]);
    i++;
  }
  
  (this->*renderpcaFunc)();
  (this->*applyColorMatrixFunc)();
  
  glEnable(GL_TEXTURE_2D);
  glEnable(GL_BLEND);
     
  glBindTexture(GL_TEXTURE_2D,tex);
  
  glTexImage2D(GL_TEXTURE_2D,0,
	       GL_RGB,largestBasisWidth,largestBasisHeight,
	       0,GL_RGBA,GL_UNSIGNED_BYTE,image);
  //printf("largestBasisWidth %d, largestBasisHeight %d\n",largestBasisWidth,
  //	 largestBasisHeight);
  /*
  inittex(image,256,256,GL_RGBA,GL_DECAL);
  glTexImage2D(GL_TEXTURE_2D,0,
	       GL_RGB,256,256,
	       0,GL_RGBA,GL_UNSIGNED_BYTE,image);
  */
  /*OLD
  glDisable(GL_BLEND); //added NB june 10 for soft
  glDisable(GL_TEXTURE_2D);//added NB june 10 for soft
  
  glPixelStorei(GL_PACK_ALIGNMENT,1);
  glRasterPos2i(0,0);
  glDrawPixels(256,256,GL_RGBA,GL_UNSIGNED_BYTE,image);
  glFlush();
  */
  printerror("PcaRendererSoft::renderpca");
  //glEnable(GL_BLEND);
  //glEnable(GL_TEXTURE_2D);
  //glutSwapBuffers();
}







