#ifndef PCA_NV_H
#define PCA_NV_H

#include "useful.h"
#include <zlib.h>
#include <string.h>

class PcaRenderer
{
 public:

  virtual ~PcaRenderer() { };
  void initNumChannels(int numChannels);
  virtual void setDisplayList(int l)=NULL;
  virtual void destroy()=NULL;
  virtual int load(char *basepath,char **filenames)=NULL;

  virtual int initBasis(int channel,int width,int height,int numbasis)=NULL;

  virtual int loadbasis(int channel,int start,int howmany,gzFile file)=NULL;
  virtual int loadbasis(int channel,char *filename)=NULL;
  virtual int loadbasis(int channel,gzFile file)=NULL;
  virtual inline void renderpca(char **coeff,GLuint tex)=NULL;
  virtual void setReduceFactor(int r)=NULL;
  virtual int getUsingHardware()=NULL;
  virtual int getUsingMMX(){return 0;}
  virtual void setUsingMMX(int i){};
  int getReduceFactor() {return reduceFactor;}
  void setQuality(float q)
    {
      int channel=0;
      //clamp the new quality
      if(q<0)
	q=0;
      if(q>1.0f)
	q=1.0f;

      quality=q;
      
      while(channel<nrchannels){
	nruse[channel]=(int)(nreig[channel]*q);
	nruse[channel]&=(~0xF);//truncate to 16
	if(nruse[channel]==0)nruse[channel]=16;
	channel++;
      }
    }
  float getQuality()
    {
      return quality;
    }
  void setChannelUse(int channel, int q)
    {
      if(channel>=0 && channel< nrchannels){
	nruse[channel]=q;
      }
    }
  int getChannelUse(int channel)
    {
      if(channel>=0 && channel< nrchannels){
	return nruse[channel];
      }
      return 0;
    }
  void setEig(int channel,int value)
    {
      if(channel>=0 && channel < nrchannels){
	nreig[channel]=value;
      }
    }
  int getEig(int channel)
    {
      return nreig[channel];
    }
  
  int nrchannels;
  int *nreig;
  int *nruse;
  int reduceFactor;
  float *colormatrix;
  float quality;

  int largestBasisWidth;
  int largestBasisHeight;
  int * basisWidth;
  int * basisHeight;
  
 protected:
  GLuint pcalist;    
  GLuint **eigentex;
};

typedef struct  rgbt
{
  unsigned char r,g,b,a;
}rgb;

class PcaRendererSoft: public PcaRenderer
{
 public:
  char ** basis; //The actual basis as chars
  char ** currentBasis;
  int * currentBasisWidth;
  int * currentBasisHeight;
  //int * basisWidth;
  //int * basisHeight;
  rgb * image;
  short ** imageShort;
  //int largestBasisWidth;
  //int largestBasisHeight;
  
  PcaRendererSoft(int num_channels=3);
  ~PcaRendererSoft()
    {
      destroy();
    }

  virtual void setDisplayList(int l)//FIXME
    {
      
    }
  virtual void setReduceFactor(int r);
  virtual int getUsingHardware(){return 0;}
  virtual int getUsingMMX(){return !(renderpcaFunc==
				     &PcaRendererSoft::renderpcaNoMMX);};
  virtual void setUsingMMX(int s){
    printDebug("in setUsingMMX %d\n",s);
    if(s==1){
      renderpcaFunc = &PcaRendererSoft::renderpcaMMX2;
      applyColorMatrixFunc=&PcaRendererSoft::applyColorMatrixMMX;
    }
    else if(s==2){
      renderpcaFunc = &PcaRendererSoft::renderpcaMMX2;
      applyColorMatrixFunc=&PcaRendererSoft::applyColorMatrixMMX;
    }
    else {
      renderpcaFunc = &PcaRendererSoft::renderpcaNoMMX;
      applyColorMatrixFunc = &PcaRendererSoft::applyColorMatrixNoMMX;
    }
  }
 protected:
  char colorMatrixShort[16];

  void (PcaRendererSoft::*renderpcaFunc)(void);
  void (PcaRendererSoft::*applyColorMatrixFunc)(void);
  void renderpcaMMX();
  void renderpcaMMX2();
  void renderpcaNoMMX();
  void applyColorMatrixNoMMX();
  void applyColorMatrixMMX();
  virtual void destroy();
  virtual int initBasis(int channel,int width,int height,int numbasis);
  virtual int load(char *basepath,char **filenames);
  virtual int loadbasis(int channel,int start,int howmany,gzFile file);
  virtual int loadbasis(int channel,gzFile file);
  virtual int loadbasis(int channel,char *filename);
  void renderpca(char ** coeff,GLuint tex);
  void renderpca(float ** coeff,GLuint tex);
  void renderpcabody(GLuint tex);
 public:
  static int numCreated;
  static int hasMMX;
  static int getMMX(void );
};

class PcaRendererNV : public PcaRenderer
{
 public:
  PcaRendererNV(int num_channels=3);
  ~PcaRendererNV();
  virtual void setDisplayList(int l);
  virtual void destroy();
  virtual int load(char *basepath,char **filenames);
  virtual int initBasis(int channel,int width,int height,int numbasis);
  virtual int loadbasis(int channel,int start,int howmany,gzFile file);
  virtual int loadbasis(int channel,gzFile  file);
  virtual int loadbasis(int channel,char *filename);
  virtual inline void renderpca(char ** coeff,GLuint tex)
    {(this->*_renderpca)(coeff,tex);}
  
  virtual void setReduceFactor(int r);
  virtual int getUsingHardware(){return 1;}
 protected:
  int largestBasisWidth;
  int largestBasisHeight;
  int reinitTexture;
  
  void (PcaRendererNV::*setup_pca_state)(void);
  void (PcaRendererNV::*_renderpca)(char **coeff,GLuint tex);

  void setup_pca_state_NV17();
  void setup_pca_state_NV20();    
  void renderpca_NV17(char **coeff,GLuint tex);
  void renderpca_NV20(char **coeff,GLuint tex);
};

//FIX: switch modes, creating new PcaRenderer
class Pca
{
 public:
  static int numCreated;
  static int hardwareSupported;
  static void initialize();
  static int getHardwareSupported();
 public:
  Pca();
  Pca(int num_channels);
  ~Pca()
    {destroy();}
  void setDisplayList(int l)
    {pcaRenderer->setDisplayList(l);}
  void destroy();

  int load(char *basepath,char **filenames)
    {pcaRenderer->load(basepath,filenames);}
  
  int initBasis(int channel,int width,int height,int numbasis){
    if(pcaRenderer){
      return pcaRenderer->initBasis(channel,width,height,numbasis);
    }
    return 0;
  }

  int loadbasis(int channel,int start,int numbasis,gzFile gzfile){
    if(pcaRenderer){
      return pcaRenderer->loadbasis(channel,start,numbasis,gzfile);
    }
    return 0;
  }
  
  int loadbasis(int channel,gzFile  gzfile){
    if(pcaRenderer){
      return pcaRenderer->loadbasis(channel,gzfile);
    }
    return 0;
  }
  /*int loadbasis(int channel,char *filename){
    if(currentFilenames[channel]){
      delete currentFilenames[channel];
      currentFilenames[channel]=NULL;
    }
    currentFilenames[channel] = new char[strlen(filename)+1];
    printf("in load %d, f[%d]=%p\n",channel,channel,
	   currentFilenames[channel]);
    
    strcpy(currentFilenames[channel],filename); 
    pcaRenderer->loadbasis(channel,filename);
    }*/
  inline void renderpca(char ** coeff,GLuint tex)
    { pcaRenderer->renderpca(coeff,tex);}
  void setReduceFactor(int r)
    {pcaRenderer->setReduceFactor(r);}
  int getReduceFactor()
    {return pcaRenderer->getReduceFactor();}
  void setQuality(float q)
    {pcaRenderer->setQuality(q);}
  float getQuality()
    {return pcaRenderer->getQuality();}
  void setChannelUse(int channel, int use){
      pcaRenderer->setChannelUse(channel,use);
  }
  int getChannelUse(int channel){
      return pcaRenderer->getChannelUse(channel);
  }
  void  setEig(int channel,int value){
      pcaRenderer->setEig(channel,value);
  }
  int getEig(int channel){
      return pcaRenderer->getEig(channel);
  }
  int setUsingHardware(int value){
    int changed=0;
    int i=0;
    PcaRenderer * oldPcaRenderer = pcaRenderer;
    if(value){
      //FIXME: make sure you can change the mode
      if(!getUsingHardware()){
	printDebug("setting to hardware mode\n");
	pcaRenderer = new PcaRendererNV(nrchannels);
	changed=1;
      }
      else
	printDebug("Already in hardware mode\n");
    }
    else{
      if(getUsingHardware()){
	printf("setting to software mode\n");
	pcaRenderer = new PcaRendererSoft(nrchannels);
	changed=1;
      }
      else printDebug("Already in software mode\n");
    }
    if(changed){
      if(oldPcaRenderer){
	pcaRenderer->setQuality(oldPcaRenderer->getQuality());
	pcaRenderer->setReduceFactor(oldPcaRenderer->getReduceFactor());
	delete oldPcaRenderer;
      }
      printDebug("finished switching the modes\n");
    }
  }
  int getUsingHardware(){
    return pcaRenderer->getUsingHardware();
  }
  int getUsingMMX(){
    return pcaRenderer->getUsingMMX();
  }
  void setUsingMMX(int val){
    if(pcaRenderer)
      pcaRenderer->setUsingMMX(val);
  }
  private:
    PcaRenderer * pcaRenderer;
    int nrchannels;
};

#endif







