#include"useful.h"
#include<GL/gl.h>
#include<stdlib.h>

#ifdef _WIN32
PFNGLACTIVETEXTUREARBPROC glActiveTextureARB;
PFNGLMULTITEXCOORD2FARBPROC glMultiTexCoord2fARB;
PFNGLMULTITEXCOORD2FVARBPROC glMultiTexCoord2fvARB;

PFNGLBLENDEQUATIONPROC glBlendEquation;
PFNGLBLENDCOLORPROC glBlendColor;

PFNGLCOMBINERPARAMETERFVNVPROC glCombinerParameterfvNV;
PFNGLCOMBINERPARAMETERIVNVPROC glCombinerParameterivNV;
PFNGLCOMBINERPARAMETERFNVPROC glCombinerParameterfNV;
PFNGLCOMBINERPARAMETERINVPROC glCombinerParameteriNV;
PFNGLCOMBINERSTAGEPARAMETERFVNVPROC glCombinerStageParameterfvNV;
PFNGLCOMBINERINPUTNVPROC glCombinerInputNV;
PFNGLCOMBINEROUTPUTNVPROC glCombinerOutputNV;
PFNGLFINALCOMBINERINPUTNVPROC glFinalCombinerInputNV;
PFNGLGETCOMBINERINPUTPARAMETERFVNVPROC glGetCombinerInputParameterfvNV;
PFNGLGETCOMBINERINPUTPARAMETERIVNVPROC glGetCombinerInputParameterivNV;
PFNGLGETCOMBINEROUTPUTPARAMETERFVNVPROC glGetCombinerOutputParameterfvNV;
PFNGLGETCOMBINEROUTPUTPARAMETERIVNVPROC glGetCombinerOutputParameterivNV;
PFNGLGETFINALCOMBINERINPUTPARAMETERFVNVPROC glGetFinalCombinerInputfvNV;
PFNGLGETFINALCOMBINERINPUTPARAMETERIVNVPROC glGetFinalCombinerInputivNV;


void initGLext()
{
  const unsigned char *s=glGetString(GL_EXTENSIONS);
  glActiveTextureARB=(PFNGLACTIVETEXTUREARBPROC)wglGetProcAddress("glActiveTextureARB");
  glMultiTexCoord2fARB=(PFNGLMULTITEXCOORD2FARBPROC)wglGetProcAddress("glMultiTexCoord2fARB");
  glMultiTexCoord2fvARB=(PFNGLMULTITEXCOORD2FVARBPROC)wglGetProcAddress("glMultiTexCoord2fvARB");

  glBlendEquation=(PFNGLBLENDEQUATIONPROC)wglGetProcAddress("glBlendEquation");
  glBlendColor=(PFNGLBLENDCOLORPROC)wglGetProcAddress("glBlendColor");

  glCombinerParameterfvNV=(PFNGLCOMBINERPARAMETERFVNVPROC)wglGetProcAddress("glCombinerParameterfvNV");
  glCombinerParameterivNV=(PFNGLCOMBINERPARAMETERIVNVPROC)wglGetProcAddress("glCombinerParameterivNV");
  glCombinerParameterfNV=(PFNGLCOMBINERPARAMETERFNVPROC)wglGetProcAddress("glCombinerParameterfNV");
  glCombinerParameteriNV=(PFNGLCOMBINERPARAMETERINVPROC)wglGetProcAddress("glCombinerParameteriNV");
  glCombinerStageParameterfvNV=(PFNGLCOMBINERSTAGEPARAMETERFVNVPROC)wglGetProcAddress("glCombinerStageParameterfvNV");
  glCombinerInputNV =(PFNGLCOMBINERINPUTNVPROC)wglGetProcAddress("glCombinerInputNV");
  glCombinerOutputNV =(PFNGLCOMBINEROUTPUTNVPROC)wglGetProcAddress("glCombinerOutputNV");
  glFinalCombinerInputNV =(PFNGLFINALCOMBINERINPUTNVPROC)wglGetProcAddress("glFinalCombinerInputNV");
  glGetCombinerInputParameterfvNV =(PFNGLGETCOMBINERINPUTPARAMETERFVNVPROC)wglGetProcAddress("glGetCombinerInputParameterfvNV");
  glGetCombinerInputParameterivNV =(PFNGLGETCOMBINERINPUTPARAMETERIVNVPROC)wglGetProcAddress("glGetCombinerInputParameterivNV");
  glGetCombinerOutputParameterfvNV =(PFNGLGETCOMBINEROUTPUTPARAMETERFVNVPROC)wglGetProcAddress("glGetCombinerOutputParameterfvNV");
  glGetCombinerOutputParameterivNV =(PFNGLGETCOMBINEROUTPUTPARAMETERIVNVPROC)wglGetProcAddress("glGetCombinerOutputParameterivNV");
  glGetFinalCombinerInputfvNV =(PFNGLGETFINALCOMBINERINPUTPARAMETERFVNVPROC)wglGetProcAddress("glGetFinalCombinerInputfvNV");
  glGetFinalCombinerInputivNV =(PFNGLGETFINALCOMBINERINPUTPARAMETERIVNVPROC)wglGetProcAddress("glGetFinalCombinerInputivNV");
}
#endif

void printDebug(char *fmt, ...)
{
#ifdef PRINT_DEBUG
  va_list args;
  
  va_start(args, fmt);

  vfprintf(stdout,fmt,args);
    
  va_end(args);
#endif
}


void printerror(char *where)
{
  int err=glGetError();
  switch(err)
  { 
    case 0: break; //not an error
    case GL_INVALID_ENUM: printf("%s: invalid enum\n",where);
			  break;
    case GL_INVALID_OPERATION: printf("%s: invalid operation\n",where);
			  break;
    case GL_INVALID_VALUE: printf("%s: invalid value\n",where);
			  break;
    default:
    printf("opengl error %d !!!\n",err);
  }
}

void inittex(void *data,int w,int h,int type,int mode)
{
  printDebug("initing new tex: %d by %d\n", w,h);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
  glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,mode);
  glTexImage2D(GL_TEXTURE_2D,0,type,w,h,0,type,GL_UNSIGNED_BYTE,data);
  printerror("inittex");
}

int writeppm(char *filename,int w,int h,char *data)
{
  FILE *f=fopen(filename,"wb");
  if(!f)return 0;
  fprintf(f,"P6\n#created by keith\n%d %d\n255\n",w,h);
  for(int i=h-1;i>=0;i--)
  {
    fwrite(data+i*w*3,w*3,1,f);
  }
  return 1;
}

char *readline(FILE *fp)
{
  int ch;
  char *s=NULL;
  int i=0;
  int size=0;

  while((ch=fgetc(fp))!='\n' && ch!='\r' && ch!=EOF)
  {
    if(i>=size) s=(char *)realloc(s,size+=8);
    s[i++]=ch;
  }
  if(ch=='\r')fgetc(fp);
  if(s!=NULL)s[i]='\0';
  else if(ch!=EOF)
  {
    s=(char *)malloc(1);
    s[0]='\0';
  }
  return s;
}

// count lines starting with str
int linecount(FILE *fd,char *str)
{
  char *line;
  int i,len=strlen(str),count=0;

  fseek(fd,0,SEEK_SET);
  while(line=readline(fd))
  {
    for(i=0;i<len && str[i]==line[i];i++);
    if(i==len)
      count++;
    delete line;
  }
  fseek(fd,0,SEEK_SET);
  return count;
}

// get next line starting with str
char *nextline(FILE *fd,char *str)
{
  char *line;
  int i,len=strlen(str);

  while(line=readline(fd))
  {
    for(i=0;i<len && str[i]==line[i];i++);
    if(i==len)break;
    delete line;
  }
  return line;
}

int gzGetLine(gzFile gz,char * str,int len)
{
  int i=0;
  int res;
  while(i<len){
    res=gzgetc(gz);
    if(res==EOF)
      return EOF;
    str[i]=res;
    i++;
    if(res=='\n')
      break;
  }
  str[i]=0;
  //printf("gzgetline=%s",str);
  return i;
}

int swaplong(int a)
{
  int b;
  char *bp1=(char *)&a;
  char *bp2=(char *)&b;
  
  bp2[0]=bp1[3];
  bp2[1]=bp1[2];
  bp2[2]=bp1[1];
  bp2[3]=bp1[0];

  return b;
}

short int swapshort(short int a)
{
  short int b;
  char *bp1=(char *)&a;
  char *bp2=(char *)&b;
  
  bp2[0]=bp1[1];
  bp2[1]=bp1[0];

  return b;
}

float swapfloat(float a)
{
  float b;
  char *bp1=(char *)&a;
  char *bp2=(char *)&b;
  
  bp2[0]=bp1[3];
  bp2[1]=bp1[2];
  bp2[2]=bp1[1];
  bp2[3]=bp1[0];

  return b;
}
