#include"useful.h"
#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<GL/gl.h>
#include<GL/glext.h>
#include"pcamesh.h"

#include <stdarg.h>


PcaMesh::PcaMesh()
{
  char *emptytex=new char[256*256*3];
  int ntexunits;
  int i=0;

  while(i<256*256*3){
    emptytex[i]=i%256;
    i++;
  }
  glGenTextures(1,&tex);
  glBindTexture(GL_TEXTURE_2D,tex);
  inittex(emptytex,256,256,GL_RGB,GL_MODULATE);
  delete emptytex;

  if(Pca::getHardwareSupported()){
    glGetIntegerv(GL_MAX_TEXTURE_UNITS_ARB,&ntexunits);
    printDebug("You have %d\n",ntexunits);
    if(ntexunits<2){
	printf("your graphics card is not good enuf. goodbye");
	exit(0);
    }
    setup_bluescreen=glGenLists(1);
    disable_combiners=glGenLists(1);
    if(ntexunits>=4){
      glNewList(setup_bluescreen,GL_COMPILE);
        setup_bluescreen_state_NV17(); 
      glEndList();
    
      glNewList(disable_combiners,GL_COMPILE);
       glDisable(GL_REGISTER_COMBINERS_NV);
       glActiveTextureARB(GL_TEXTURE1_ARB);
       glDisable(GL_TEXTURE_2D);
       glActiveTextureARB(GL_TEXTURE2_ARB);
       glDisable(GL_TEXTURE_2D);
       glActiveTextureARB(GL_TEXTURE3_ARB);
       glDisable(GL_TEXTURE_2D);
       glActiveTextureARB(GL_TEXTURE0_ARB);
      glEndList();
    }
    else{
      glNewList(setup_bluescreen,GL_COMPILE);
	setup_bluescreen_state_NV17();
      glEndList();
      
      glNewList(disable_combiners,GL_COMPILE);
       glDisable(GL_REGISTER_COMBINERS_NV);
       glActiveTextureARB(GL_TEXTURE1_ARB);
       glDisable(GL_TEXTURE_2D);
       glActiveTextureARB(GL_TEXTURE0_ARB);
      glEndList();
    }
  }
  else {
    setup_bluescreen=glGenLists(1);
    glNewList(setup_bluescreen,GL_COMPILE);
    glEndList();
    
    disable_combiners=glGenLists(1);
    
    glNewList(disable_combiners,GL_COMPILE);
    glEndList();
  }
  memset(lut,0,sizeof(float *)*3);
  memset(coefficients,0,sizeof(float *)*3);
}

PcaMesh::~PcaMesh()
{
  destroy();
}

void PcaMesh::destroy()
{
  printDebug("destroying pcamesh\n");

  glDeleteLists(setup_bluescreen,1);
  glDeleteLists(disable_combiners,1);
  
  int i=0;
  while(i<3){
    if(lut[i])
      delete lut[i];
    if(coefficients[i])
      delete coefficients[i];
    i++;
  }
}

/*
  post: load in the entire directory structure specified
  by path.  This method only exists for backward compatibility
*/
int PcaMesh::loadDir(char *path)
{
  char s[256],t[256];
  char *line;
  FILE *fd;
  gzFile  gzf;
  gzFile  gzf2;

  // load geometry
  strncpy(s,path,256);
  strncat(s,"/geometry.obj",256);
  fd=fopen(s,"r");
  if(!loadobj(fd))
  {
    destroy();
    fclose(fd);
    return 0;
  }

  fseek(fd,0,SEEK_SET);
  if(line=nextline(fd,"mean_x ")){
    sscanf(line,"mean_x %f %f %f",mean_X,mean_X+1,mean_X+2);
    eul[0]=mean_X[0]; eul[1]=mean_X[1]; eul[2]=mean_X[2];
  }
  fseek(fd,0,SEEK_SET);
  if(line=nextline(fd,"min_x "))
    sscanf(line,"min_x %f %f %f",min_X,min_X+1,min_X+2);
  fseek(fd,0,SEEK_SET);
  if(line=nextline(fd,"max_x "))
    sscanf(line,"max_x %f %f %f",max_X,max_X+1,max_X+2);

  fclose(fd);
  
  strncpy(s,path,256);
  strncat(s,"/gridY",256);
  strncpy(t,path,256);
  strncat(t,"/dataY",256);

  if(!((gzf=gzopen(s,"rb"))&&(gzf2=gzopen(t,"rb")))){
    destroy();
    return 0;
  }
  if(!(loadlut(0,gzf) && loadbasis(0,gzf2)))
  {
    destroy();
    return 0;
  }
  gzclose(gzf);gzclose(gzf2);

  s[strlen(s)-1]='U';
  t[strlen(t)-1]='U';
  if(!((gzf=gzopen(s,"rb"))&&(gzf2=gzopen(t,"rb")))){
    destroy();
    return 0;
  }
  if(!(loadlut(1,gzf) && loadbasis(1,gzf2)))
  {
    destroy();
    return 0;
  }
  gzclose(gzf);gzclose(gzf2);

  s[strlen(s)-1]='V';
  t[strlen(t)-1]='V';
  if(!((gzf=gzopen(s,"rb"))&&(gzf2=gzopen(t,"rb")))){
    destroy();
    return 0;
  }
  if(!(loadlut(2,gzf) && loadbasis(2,gzf2)))
  {
    destroy();
    return 0;
  }
  gzclose(gzf);gzclose(gzf2);
  
  //nruse[1]=nruse[2]=16; rem jun 9 nb
  setChannelUse(1,16);//added jun 9
  setChannelUse(2,16);//..
}

/*
  post: initialize the width and height of the lookup table,
  and allocate the data.  Must be called before any loading is
  done
*/
int PcaMesh::initlut(int channel,int width,int height)
{
  int tempint = getEig(channel);
  lutwidth[channel]=width;
  lutheight[channel]=height;

  if(lut[channel])
    delete lut[channel];
  if(coefficients[channel])
    delete coefficients[channel];
 
  lut[channel]=new char[lutwidth[channel]*lutheight[channel]*tempint];
  coefficients[channel]=new char[tempint+1];

  memset(lut[channel],0,lutwidth[channel]*lutheight[channel]*tempint);
  memset(coefficients[channel],0,tempint+1);

  return 1;
}

/*
  post: load into the channel the specific number of basis' lookup tables starting
  at start.  If usingbytes is not set, the incoming data is floating point, and must be 
  scaled to fit into a byte
*/
int PcaMesh::loadlut(int channel,int start,int howmany,gzFile f,int usingbytes)
{
  int tempint=getEig(channel);

  printDebug("loadlut %d width=%d height=%d\n",channel,
	 lutwidth[channel],lutheight[channel]);
  printDebug("loadlut: start=%d howmany=%d %d\n",start,
	 howmany,
	 sizeof(float)*lutwidth[channel]
	 *lutheight[channel]*howmany);

  /*
    We are receiving bytes, so just read them directly into the
    buffer
  */
  if(usingbytes){  
    if(gzread(f,lut[channel]+start*lutwidth[channel]*lutheight[channel]
	      ,sizeof(unsigned char)*lutwidth[channel]
	      *lutheight[channel]*howmany)<0)
      return 0;
  }
  /*
    We are receiving floats, so read them into a temp buffer, and
    convert them allinto the byte format
  */
  else{
    float * floats = new float[lutwidth[channel]*lutheight[channel]*howmany];
    char * lutstart = lut[channel]+start*lutwidth[channel]*lutheight[channel];

    if(gzread(f,floats
	      ,sizeof(float)*lutwidth[channel]
	      *lutheight[channel]*howmany)<0)
      return 0;
    
    /*
      Scale the floats into bytes
    */
    int i=0;
    while(i<lutwidth[channel]
	  *lutheight[channel]*howmany){
      lutstart[i] = (char)(floats[i]*127.0f);//-1 becomes -127 and 1 --> 127
      i++;
    }
    delete floats;
  }

  printDebug("done loading lut\n");
  printerror("loadlut");
  return 1;
}

/*
  post: this is the old loadlut for loading a directory.  It makes use
  of the newer initLut and loadlut above.  The file has both the widht 
  and height and nreig, which must be stripped before calling the
  new loadlut
*/
int PcaMesh::loadlut(int channel,gzFile  f)
{
  int width,height;

  printDebug("call to loadlut is deprecated.");

  if(gzread(f,&width,4)<0)
    return 0;
  if(gzread(f,&height,4)<0)
    return 0;

  int tempint;
  if(gzread(f,&tempint,4)<0)
    return 0;

  setEig(channel,tempint);

  initlut(channel,width,height);
  loadlut(channel,0,tempint,f,0);

  printDebug("done loading lut\n");
  printerror("loadlut");

  return 1;
}

void PcaMesh::updatetexture()
{
  if(lut[0])
    updatetexture(eul);
}

void PcaMesh::updatetexture(float *a)
{
  float angle[3]={a[0],a[1],a[2]};
  int i,channel,xind,yind;
  
  if(angle[0]>max_X[0])angle[0]=max_X[0];
  if(angle[0]<min_X[0])angle[0]=min_X[0];			       
  if(angle[1]>max_X[1])angle[1]=max_X[1];
  if(angle[1]<min_X[1])angle[1]=min_X[1];
  if(angle[2]>max_X[2])angle[2]=max_X[2];
  if(angle[2]<min_X[2])angle[2]=min_X[2];

  
  for(channel=0;channel<3;channel++){
    int nuse = getChannelUse(channel);
    xind=(int)((lutwidth[channel]-1)
	       *(angle[1]-min_X[1])/(max_X[1]-min_X[1]));
    yind=(int)((lutheight[channel]-1)
	       *(angle[2]-min_X[2])/(max_X[2]-min_X[2]));
    
    if(xind<0)xind=0; if(xind>=lutwidth[channel])xind=lutwidth[channel]-1;
    if(yind<0)yind=0; if(yind>=lutheight[channel])yind=lutheight[channel]-1;
    
    coefficients[channel][0]=127;    
    for(int i=0;i<(nuse-1);i++){
      coefficients[channel][i+1]=
	lut[channel][lutwidth[channel]*lutheight[channel]*i + 
		      yind*lutwidth[channel]+xind];
    }
  }
  
/* OLD FOR FLOATS
    for(channel=0;channel<3;channel++){
      int nuse = getChannelUse(channel);
      xind=(int)((lutwidth[channel]-1)*(angle[1]-min_X[1])/(max_X[1]-min_X[1]));
      yind=(int)((lutheight[channel]-1)*(angle[2]-min_X[2])/(max_X[2]-min_X[2]));
      
      if(xind<0)xind=0; if(xind>=lutwidth[channel])xind=lutwidth[channel]-1;
      if(yind<0)yind=0; if(yind>=lutheight[channel])yind=lutheight[channel]-1;
      
      coefficients[channel][0]=1;    
      for(int i=0;i<(nuse-1);i++)
	{
	  coefficients[channel][i+1]=
	    lut[channel][lutwidth[channel]*lutheight[channel]*i + 
			yind*lutwidth[channel]+xind];
	}
    }
*/
  //(this->*_renderpca)(coefficients); 
  renderpca(coefficients,tex);//added june 9
}

void PcaMesh::setup_bluescreen_state_NV20()
{
  glEnable(GL_REGISTER_COMBINERS_NV);  
  glCombinerParameteriNV(GL_NUM_GENERAL_COMBINERS_NV,3);

  glCombinerInputNV(GL_COMBINER0_NV, GL_RGB, GL_VARIABLE_A_NV, GL_ZERO, GL_UNSIGNED_INVERT_NV, GL_RGB); 
  glCombinerInputNV(GL_COMBINER0_NV, GL_RGB, GL_VARIABLE_B_NV, GL_TEXTURE0,    GL_SIGNED_IDENTITY_NV,   GL_RGB);
  glCombinerInputNV(GL_COMBINER0_NV, GL_RGB, GL_VARIABLE_C_NV, GL_ZERO, GL_UNSIGNED_INVERT_NV, GL_RGB);
  glCombinerInputNV(GL_COMBINER0_NV, GL_RGB, GL_VARIABLE_D_NV, GL_PRIMARY_COLOR_NV, GL_SIGNED_NEGATE_NV, GL_RGB);

  glCombinerOutputNV(GL_COMBINER0_NV, GL_RGB, GL_DISCARD_NV, GL_DISCARD_NV, GL_SPARE0_NV,GL_NONE,GL_NONE,GL_FALSE,GL_FALSE, GL_FALSE);
  
  printerror("rc0 -2");

  glCombinerInputNV(GL_COMBINER1_NV, GL_RGB, GL_VARIABLE_A_NV, GL_SPARE0_NV,GL_SIGNED_IDENTITY_NV, GL_RGB); 
  glCombinerInputNV(GL_COMBINER1_NV, GL_RGB, GL_VARIABLE_B_NV, GL_SPARE0_NV,GL_SIGNED_IDENTITY_NV,   GL_RGB);
 
  glCombinerOutputNV(GL_COMBINER1_NV, GL_RGB, GL_SPARE0_NV, GL_DISCARD_NV, GL_DISCARD_NV,GL_NONE,GL_NONE,GL_TRUE,GL_FALSE, GL_FALSE);

  printerror("rc1 -2");

  glCombinerInputNV(GL_COMBINER2_NV, GL_RGB, GL_VARIABLE_A_NV, GL_SPARE0_NV,GL_SIGNED_IDENTITY_NV, GL_RGB); 
  glCombinerInputNV(GL_COMBINER2_NV, GL_RGB, GL_VARIABLE_B_NV, GL_SPARE0_NV,GL_SIGNED_IDENTITY_NV,   GL_RGB);
  glCombinerInputNV(GL_COMBINER2_NV, GL_RGB, GL_VARIABLE_C_NV, GL_ZERO,GL_UNSIGNED_IDENTITY_NV,   GL_RGB);
  glCombinerInputNV(GL_COMBINER2_NV, GL_RGB, GL_VARIABLE_D_NV, GL_ZERO,GL_UNSIGNED_IDENTITY_NV,   GL_RGB);
 
  glCombinerOutputNV(GL_COMBINER2_NV, GL_RGB, GL_DISCARD_NV, GL_DISCARD_NV, GL_SPARE0_NV,GL_SCALE_BY_TWO_NV,GL_NONE,GL_FALSE,GL_FALSE, GL_FALSE);

  printerror("rc2 -2");

  glFinalCombinerInputNV(GL_VARIABLE_A_NV,GL_ZERO, GL_UNSIGNED_INVERT_NV, GL_RGB);
  glFinalCombinerInputNV(GL_VARIABLE_B_NV,GL_TEXTURE0,GL_UNSIGNED_IDENTITY_NV, GL_RGB);
  glFinalCombinerInputNV(GL_VARIABLE_C_NV,GL_ZERO, GL_UNSIGNED_IDENTITY_NV, GL_RGB);
  glFinalCombinerInputNV(GL_VARIABLE_D_NV,GL_ZERO, GL_UNSIGNED_IDENTITY_NV, GL_RGB);
  glFinalCombinerInputNV(GL_VARIABLE_G_NV,GL_SPARE0_NV, GL_UNSIGNED_IDENTITY_NV, GL_BLUE);

  printerror("rcF -2");
}

void PcaMesh::setup_bluescreen_state_NV17()
{
  glEnable(GL_REGISTER_COMBINERS_NV);
  glCombinerParameteriNV(GL_NUM_GENERAL_COMBINERS_NV,2);
  glCombinerInputNV(GL_COMBINER0_NV, GL_RGB, GL_VARIABLE_A_NV, GL_ZERO, GL_UNSIGNED_INVERT_NV, GL_RGB); 
  glCombinerInputNV(GL_COMBINER0_NV, GL_RGB, GL_VARIABLE_B_NV, GL_TEXTURE0,    GL_SIGNED_IDENTITY_NV,   GL_RGB);
  glCombinerInputNV(GL_COMBINER0_NV, GL_RGB, GL_VARIABLE_C_NV, GL_ZERO, GL_UNSIGNED_INVERT_NV, GL_RGB);
  glCombinerInputNV(GL_COMBINER0_NV, GL_RGB, GL_VARIABLE_D_NV, GL_PRIMARY_COLOR_NV, GL_SIGNED_NEGATE_NV, GL_RGB);

  glCombinerOutputNV(GL_COMBINER0_NV, GL_RGB, GL_DISCARD_NV, GL_DISCARD_NV, GL_SPARE0_NV,GL_NONE,GL_NONE,GL_FALSE,GL_FALSE, GL_FALSE);
  
  printerror("rc0 -2");

  glCombinerInputNV(GL_COMBINER1_NV, GL_RGB, GL_VARIABLE_A_NV, GL_SPARE0_NV,GL_SIGNED_IDENTITY_NV, GL_RGB); 
  glCombinerInputNV(GL_COMBINER1_NV, GL_RGB, GL_VARIABLE_B_NV, GL_SPARE0_NV,GL_SIGNED_IDENTITY_NV,   GL_RGB);
 
  glCombinerOutputNV(GL_COMBINER1_NV, GL_RGB, GL_SPARE0_NV, GL_DISCARD_NV, GL_DISCARD_NV,GL_SCALE_BY_ONE_HALF_NV,GL_NONE,GL_TRUE,GL_FALSE, GL_FALSE);

  printerror("rc1 -2");

  glFinalCombinerInputNV(GL_VARIABLE_A_NV,GL_ZERO, GL_UNSIGNED_INVERT_NV, GL_RGB);
  glFinalCombinerInputNV(GL_VARIABLE_B_NV,GL_TEXTURE0,GL_UNSIGNED_IDENTITY_NV, GL_RGB);
  glFinalCombinerInputNV(GL_VARIABLE_C_NV,GL_ZERO, GL_UNSIGNED_IDENTITY_NV, GL_RGB);
  glFinalCombinerInputNV(GL_VARIABLE_D_NV,GL_ZERO, GL_UNSIGNED_IDENTITY_NV, GL_RGB);
  //glFinalCombinerInputNV(GL_VARIABLE_G_NV,GL_SPARE0_NV, GL_UNSIGNED_IDENTITY_NV, GL_BLUE);
  glFinalCombinerInputNV(GL_VARIABLE_G_NV,GL_ZERO, GL_UNSIGNED_INVERT_NV, GL_BLUE);

  printerror("rcF -2");
}

void PcaMesh::draw()
{
  //glActiveTextureARB(GL_TEXTURE0);
  glBindTexture(GL_TEXTURE_2D,tex);
  
  glCallList(setup_bluescreen);
  Mesh::draw();
  glCallList(disable_combiners);
}




